#!/usr/bin/env python

#******************************************************************************
#
#  rpnLexicographic.py
#
#  rpnChilada lexicography operators
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import itertools
import string

from mpmath import arange, fabs, fadd, ceil, floor, fmod, fmul, fneg, fprod, fsub, fsum, log10, \
                   mp, mpf, mpmathify, nint, power

from rpn.rpnBase import convertToBaseN, getBaseKDigits
from rpn.rpnFactor import getFactors
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMath import isDivisible, getPowMod
from rpn.rpnPersistence import cachedFunction
from rpn.rpnPrimeUtils import isPrime
from rpn.rpnSettings import setAccuracy
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator, getMPFIntegerAsString, \
                         listArgFunctionEvaluator
from rpn.rpnValidator import argValidator, IntValidator, ListValidator, RealValidator


#******************************************************************************
#
#  splitNumberByDigits
#
#  This splits the number into 2 numbers by splitting off half the digits
#  for the first and the rest of the digits for the second.  If there is an
#  odd number of digits then the extra digit goes to the second number.
#
#******************************************************************************

def splitNumberByDigits( n ):
    n = getMPFIntegerAsString( n )

    split = len( n ) // 2

    return mpmathify( n[ : split ] ), mpmathify( n[ split : ] )


#******************************************************************************
#
#  getDigitList
#
#******************************************************************************

def getDigitList( n, dropZeroes=False ):
    result = [ ]

    for c in getMPFIntegerAsString( n ):
        if dropZeroes and c == '0':
            continue

        result.append( int( c ) )

    return result


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def getDigitsOperator( n ):
    return getDigitList( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def getNonzeroDigitsOperator( n ):
    return getDigitList( n, dropZeroes=True )


#******************************************************************************
#
#  getDecimalDigitListOperator
#
#******************************************************************************

def getDecimalDigitList( n, k ):
    result = [ ]

    setAccuracy( k )

    digits = floor( log10( n ) )

    if digits < 0:
        for _ in arange( fsub( fabs( digits ), 1 ) ):
            result.append( 0 )

        k = fsub( k, fsub( fabs( digits ), 1 ) )

    value = fmul( n, power( 10, fsub( k, fadd( digits, 1 ) ) ) )

    for c in getMPFIntegerAsString( floor( value ) ):
        result.append( int( c ) )

    return result


@twoArgFunctionEvaluator( )
@argValidator( [ RealValidator( ), IntValidator( 1 ) ] )
def getDecimalDigitsOperator( n, k ):
    return getDecimalDigitList( n, k )


#******************************************************************************
#
#  getRightDigitsOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 1 ) ] )
def getRightDigitsOperator( n, k ):
    return fmod( n, pow( 10, k ) )


#******************************************************************************
#
#  getLeftDigitsOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 1 ) ] )
def getLeftDigitsOperator( n, k ):
    return combineDigits( getDigitList( n )[ : int( k ) ] )


#******************************************************************************
#
#  isBaseKPandigitalOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 2 ) ] )
def isBaseKPandigitalOperator( n, base ):
    digits = convertToBaseN( n, base, outputBaseDigits=True )

    for i in arange( min( int( base ), len( digits ) ) ):
        try:
            digits.index( i )
        except ValueError:
            return 0

    return 1


#******************************************************************************
#
#  sumDigitsOperator
#
#******************************************************************************

def sumDigits( n ):
    result = 0

    for c in getMPFIntegerAsString( n ):
        result = fadd( result, int( c ) )

    return result


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def sumDigitsOperator( n ):
    return sumDigits( n )


#******************************************************************************
#
#  multiplyDigitListOperator
#
#******************************************************************************

def multiplyDigitList( n, exponent = 1, dropZeroes = False ):
    if exponent < 1:
        raise ValueError( 'multiplyDigitList( ) expects a positive integer for \'exponent\'' )

    if exponent == 1:
        return fprod( getDigitList( n, dropZeroes ) )

    return fprod( [ power( i, exponent ) for i in getDigitList( n, dropZeroes ) ] )


def multiplyDigits( n ):
    return multiplyDigitList( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def multiplyDigitsOperator( n ):
    return multiplyDigitList( n )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 0 ) ] )
def multiplyDigitPowersOperator( n, k ):
    return multiplyDigitList( n, k )


def multiplyNonzeroDigits( n ):
    return multiplyDigitList( n, dropZeroes = True )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def multiplyNonzeroDigitsOperator( n ):
    return multiplyDigitList( n, dropZeroes = True )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 0 ) ] )
def multiplyNonzeroDigitPowersOperator( n, k ):
    return multiplyDigitList( n, k, dropZeroes = True )


#******************************************************************************
#
#  appendDigits
#
#******************************************************************************

def appendDigits( n, digits, digitCount ):
    if ( n ) < 0:
        return nint( fsub( fmul( floor( n ), power( 10, digitCount ) ), digits ) )

    return nint( fadd( fmul( floor( n ), power( 10, digitCount ) ), digits ) )


#******************************************************************************
#
#  addDigitsOperator
#
#******************************************************************************

def addDigits( n, k ):
    digits = int( k )

    if digits == 0:
        digitCount = 1
    else:
        digitCount = int( fadd( floor( log10( k ) ), 1 ) )

    return appendDigits( n, digits, digitCount )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 0 ) ] )
def addDigitsOperator( n, k ):
    return addDigits( n, k )


#******************************************************************************
#
#  combineDigits
#
#******************************************************************************

def combineDigits( n ):
    if isinstance( n, RPNGenerator ):
        return combineDigits( list( n ) )

    result = 0

    listResult = False

    for i in n:
        if isinstance( i, ( list, RPNGenerator ) ) and result == 0:
            listResult = True
            result = [ combineDigits( i ) ]
        elif listResult:
            if not isinstance( result, list ):
                result = [ result ]

            result.append( combineDigits( i ) )
        else:
            result = addDigits( result, i )

    return result


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def combineDigitsOperator( n ):
    return combineDigits( n )


#******************************************************************************
#
#  replaceDigitsOperator
#
#******************************************************************************

@argValidator( [ IntValidator( ), IntValidator( 0 ), IntValidator( 0 ) ] )
def replaceDigitsOperator( n, source, replace ):
    n = getMPFIntegerAsString( n )

    if source < 0:
        raise ValueError( 'source value must be a positive integer' )

    if replace < 0:
        raise ValueError( 'replace value must be a positive integer' )

    return mpmathify( n.replace( str( int( source ) ), str( int( replace ) ) ) )


#******************************************************************************
#
#  duplicateDigitsOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 0 ) ] )
def duplicateDigitsOperator( n, k ):
    return appendDigits( n, fmod( n, power( 10, nint( floor( k ) ) ) ), k )


#******************************************************************************
#
#  duplicateNumberOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 0 ) ] )
def duplicateNumberOperator( n, k ):
    if k == 0:
        return 0

    return mpmathify( getMPFIntegerAsString( n ) * int( k ) )


#******************************************************************************
#
#  reverseDigitsOperator
#
#******************************************************************************

def reverseDigits( n ):
    return mpmathify( getMPFIntegerAsString( n )[ : : -1 ] )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def reverseDigitsOperator( n ):
    return reverseDigits( n )


#******************************************************************************
#
#  isPalindromeOperator
#
#******************************************************************************

def isPalindrome( n ):
    result = getMPFIntegerAsString( n )

    length = len( result )

    for i in range( 0, length // 2 ):
        if result[ i ] != result[ -( i + 1 ) ]:
            return 0

    return 1


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isPalindromeOperator( n ):
    return isPalindrome( n )


#******************************************************************************
#
#  isPandigitalOperator
#
#******************************************************************************

def isPandigital( n ):
    n = getMPFIntegerAsString( n )

    length = len( n )

    if length < 10:
        return 0

    digitsToCheck = string.digits

    for c in digitsToCheck:
        if c not in n:
            return 0

    return 1


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isPandigitalOperator( n ):
    return isPandigital( n )


#******************************************************************************
#
#  containsDigitsOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 0 ) ] )
def containsDigitsOperator( n, k ):
    n = getMPFIntegerAsString( n )

    if isinstance( k, ( mpf, int, float ) ):
        k = getMPFIntegerAsString( k )

    for c in k:
        if c not in n:
            return 0

    return 1


#******************************************************************************
#
#  containsAnyDigitsOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 0 ) ] )
def containsAnyDigitsOperator( n, k ):
    n = getMPFIntegerAsString( n )

    if isinstance( k, ( mpf, int, float ) ):
        k = getMPFIntegerAsString( k )

    for c in k:
        if c in n:
            return 1

    return 0


#******************************************************************************
#
#  containsOnlyDigitsOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 0 ) ] )
def containsOnlyDigitsOperator( n, k ):
    if isinstance( k, ( mpf, int, float ) ):
        k = getMPFIntegerAsString( k )

    for c in getMPFIntegerAsString( n ):
        if c not in k:
            return 0

    return 1


#******************************************************************************
#
#  countDigitsOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 0 ) ] )
def countDigitsOperator( n, k ):
    n = getMPFIntegerAsString( n )

    if isinstance( k, ( mpf, int, float ) ):
        #k = set( [ char for char in getMPFIntegerAsString( k ) ] )
        k = set( getMPFIntegerAsString( k ) )
        #print( k )

    result = 0

    for c in k:
        result += n.count( c )

    return result


#******************************************************************************
#
#  getDigitCountOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def getDigitCountOperator( n ):
    if n == 0:
        return 1

    return floor( fadd( log10( fabs( n ) ), 1 ) )


#******************************************************************************
#
#  countDifferentDigitsOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def countDifferentDigitsOperator( n ):
    return len( list( set( getMPFIntegerAsString( n ) ) ) )


#******************************************************************************
#
#  getNthReversalAdditionOperator
#
#  https://en.wikipedia.org/wiki/Lychrel_number
#
#******************************************************************************

def getNthReversalAdditionGenerator( n, k ):
    nextOne = int( n  )
    yield nextOne

    for _ in arange( k ):
        nextOne = fadd( reverseDigits( nextOne ), nextOne )

        yield nextOne

        if isPalindrome( nextOne ):
            break


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 1 ) ] )
def getNthReversalAdditionOperator( n, k ):
    return RPNGenerator( getNthReversalAdditionGenerator( n, k ) )


#******************************************************************************
#
#  findPalindromeOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 1 ) ] )
def findPalindromeOperator( n, k ):
    nextOne = int( n )

    for i in range( int( k ) + 1 ):
        nextOne = reverseDigits( nextOne ) + nextOne

        if isPalindrome( nextOne ):
            return [ i + 1, nextOne ]

    return [ k, 0 ]


#******************************************************************************
#
#  isNarcissisticOperator
#
#******************************************************************************

def isNarcissistic( n ):
    digits = getDigitList( n )
    count = len( digits )

    total = 0

    for digit in digits:
        total = fadd( total, power( digit, count ) )

        if total > n:
            return 0

    return 1 if total == n else 0


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isNarcissisticOperator( n ):
    return isNarcissistic( n )


#******************************************************************************
#
#  isPerfectDigitalInvariantOperator
#
#******************************************************************************

def isPerfectDigitalInvariant( n ):
    digits = getDigitList( n )

    exponent = 1

    oldTotal = 0

    while True:
        total = 0

        for digit in digits:
            total = fadd( total, power( digit, exponent ) )

            if total > n:
                return 0

            if total == n:
                return 1

        if total == oldTotal:
            return 0

        oldTotal = total

        exponent += 1


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isPerfectDigitalInvariantOperator( n ):
    return isPerfectDigitalInvariant( n )


#******************************************************************************
#
#  isBaseKNarcissisticOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 2 ) ] )
def isBaseKNarcissisticOperator( n, k ):
    digits = getBaseKDigits( n, k )

    count = len( digits )

    total = 0

    for digit in digits:
        total = fadd( total, power( digit, count ) )

        if total > n:
            return 0

    return 1 if total == n else 0


#******************************************************************************
#
#  isGeneralizedDudeneyNumberOperator
#
#******************************************************************************

@cachedFunction( 'generalized_dudeney' )
def isGeneralizedDudeneyNumber( base, exponent ):
    '''http://www.jakob.at/steffen/dudeney.html'''
    precision = fadd( fmul( base, exponent ), 2 )

    if mp.dps < precision:
        mp.dps = precision

    n = power( base, exponent )
    return 1 if sumDigits( n ) == base else 0


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 0 ) ] )
def isGeneralizedDudeneyNumberOperator( base, exponent ):
    return isGeneralizedDudeneyNumber( base, exponent )


#******************************************************************************
#
#  isPerfectDigitToDigitInvariantOperator
#
#******************************************************************************

def isPerfectDigitToDigitInvariant( n, k ):
    '''https://en.wikipedia.org/wiki/Perfect_digit-to-digit_invariant'''
    digits = getBaseKDigits( n, k )

    total = 0

    for digit in digits:
        digit = int( digit )

        if digit != 0:
            total = fadd( total, power( digit, digit ) )

        if total > n:
            return 0

    return 1 if total == n else 0


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 0 ) ] )
def isPerfectDigitToDigitInvariantOperator( n, k ):
    return isPerfectDigitToDigitInvariant( n, k )


#******************************************************************************
#
#  isSumProductNumberOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 2 ) ] )
def isSumProductNumberOperator( n, k ):
    '''https://en.wikipedia.org/wiki/Sum-product_number'''
    digits = getBaseKDigits( n, k )
    total = fmul( fsum( digits ), fprod( digits ) )
    return 1 if total == n else 0


#******************************************************************************
#
#  isHarshadNumberOperator
#
#******************************************************************************

def isHarshadNumber( n, k ):
    '''https://en.wikipedia.org/wiki/Harshad_number'''
    digits = getBaseKDigits( n, k )
    return 1 if isDivisible( n, fsum( digits ) ) else 0


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 2 ) ] )
def isHarshadNumberOperator( n, k ):
    return isHarshadNumber( n, k )


#******************************************************************************
#
#  isKaprekarNumberOperator
#
#******************************************************************************

def isKaprekarNumber( n ):
    if n == 1:
        return 1

    if n < 9:
        return 0

    a, b = splitNumberByDigits( power( n, 2 ) )

    return 1 if ( fadd( a, b ) == n ) else 0


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isKaprekarNumberOperator( n ):
    return isKaprekarNumber( n )


#******************************************************************************
#
#  convertStringsToNumbers
#
#******************************************************************************

def convertStringsToNumbers( values ):
    for i in values:
        yield mpmathify( i )


#******************************************************************************
#
#  buildNumbersOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def buildNumbersOperator( expression ):
    digitLists = parseNumbersExpression( expression )

    if len( digitLists ) == 1:
        return RPNGenerator.createGenerator( convertStringsToNumbers, digitLists )

    return RPNGenerator.createStringProduct( digitLists )


#******************************************************************************
#
#  buildLimitedDigitNumbers
#
#******************************************************************************

def buildLimitedDigitNumbers( digits, minLength, maxLength ):
    if minLength < 1:
        raise ValueError( 'minimum length must be greater than 0' )

    if maxLength < minLength:
        raise ValueError( 'maximum length must be greater than or equal to minimum length' )

    for i in range( minLength, maxLength + 1 ):
        for item in itertools.product( digits, repeat=i ):
            number = ''.join( item )

            yield number


#******************************************************************************
#
#  parseNumbersExpression
#
#  D[D]...
#
#  One or more digit expressions where each digit expression can be:
#      - A single digit, "0" through "9"
#      - "d" - digit (equivalent to "[0-9]"
#      - "[I[I]...]" - any of the individual digit expressions
#      - "[I[I]...:m]" - permutations of any digits I, m digits long
#      - "[I[I]...:n:m]" - permutations of any digits I, from a minimum of n,
#                          up to a maximum of m digits long
#      - "s" - the designation for a "step digit", which can be one greater
#              or one less than the previous digit.  For the purposes of step
#
#
#  A digit expression (I) can be a single digit or a range.  A single digit is
#  anything from "0" through "9".  A digit range looks like: "a-b" where a
#  is a digit that is smaller than b.
#
#******************************************************************************

def parseNumbersExpression( arg ):
    if not isinstance( arg, str ):
        arg = str( arg )

    result = [ ]

    digitRange = [ '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' ]
    oddRange = [ '1', '3', '5', '7', '9' ]
    evenRange = [ '0', '2', '4', '6', '8' ]

    defaultState = 0
    startDigitState = 1
    digitState = 2
    startRangeState = 3
    numberLengthRange1 = 4
    numberLengthRange2 = 5

    state = defaultState

    currentGroup = set( )

    lengthRangeLow = ''
    lengthRangeHigh = ''
    lastDigit = ''

    for char in arg:
        #print( 'state', state, 'ch', ch, 'result', result, 'currentGroup', currentGroup )
        if state == defaultState:
            if char == 'd':
                result.append( digitRange )
            elif char == 'e':
                result.append( evenRange )
            elif char == 'o':
                result.append( oddRange )
            elif char == '[':
                state = digitState
                currentGroup = set( )
            elif '0' <= char <= '9':
                result.append( [ char ] )
            else:
                raise ValueError( 'unexpected character \'{}\''.format( char ) )
        elif state == startDigitState:
            if '0' <= char <= '9':
                currentGroup.add( char )
                state = digitState
                lastDigit = char
            elif char == ']':
                result.append( sorted( list( currentGroup ) ) )
                state = defaultState
            else:
                raise ValueError( 'unexpected character \'{}\''.format( char ) )
        elif state == digitState:
            if '0' <= char <= '9':
                currentGroup.add( char )
                lastDigit = char
            elif char == 'd':
                for digit in digitRange:
                    currentGroup.add( digit )

                lastDigit = ' '
            elif char == 'e':
                for digit in evenRange:
                    currentGroup.add( digit )

                lastDigit = ' '
            elif char == 'o':
                for digit in oddRange:
                    currentGroup.add( digit )

                lastDigit = ' '
            elif char == '-':
                state = startRangeState
            elif char == ':':
                state = numberLengthRange1
            elif char == ']':
                result.append( sorted( list( currentGroup ) ) )
                state = defaultState
        elif state == startRangeState:
            if lastDigit == ' ':
                raise ValueError( 'invalid digit range' )

            if '0' <= char <= '9':
                if char <= lastDigit:
                    raise ValueError( 'invalid digit range' )

                for i in range( int( lastDigit ), int( char ) + 1 ):
                    currentGroup.add( str( i ) )

                state = digitState
        elif state == numberLengthRange1:
            if '0' <= char <= '9':
                lengthRangeHigh += char
            elif char == ':':
                lengthRangeLow = lengthRangeHigh
                lengthRangeHigh = ''
                state = numberLengthRange2
            elif char == ']':
                result.append( buildLimitedDigitNumbers( sorted( list( currentGroup ) ),
                                                         int( lengthRangeHigh ), int( lengthRangeHigh ) ) )
                lengthRangeLow = ''
                lengthRangeHigh = ''
                state = defaultState
            else:
                raise ValueError( 'unexpected character \'{}\''.format( char ) )
        elif state == numberLengthRange2:
            if '0' <= char <= '9':
                lengthRangeHigh += char
            elif char == ']':
                result.append( buildLimitedDigitNumbers( sorted( list( currentGroup ) ),
                                                         int( lengthRangeLow ), int( lengthRangeHigh ) ) )
                lengthRangeLow = ''
                lengthRangeHigh = ''
                state = defaultState
            else:
                raise ValueError( 'unexpected character \'{}\''.format( char ) )

    return result


#******************************************************************************
#
#  hasUniqueDigitsOperator
#
#******************************************************************************

def hasUniqueDigits( n ):
    digits = getDigitList( n )

    existing = set( )

    for digit in digits:
        if digit in existing:
            return 0

        existing.add( digit )

    return 1


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def hasUniqueDigitsOperator( n ):
    return hasUniqueDigits( n )


#******************************************************************************
#
#  hasDigits
#
#******************************************************************************

#@twoArgFunctionEvaluator( )
#def hasDigits( value, digits ):
#    pass


#******************************************************************************
#
#  isKMorphicOperator
#
#******************************************************************************

def isKMorphic( n, k ):
    '''
    Returns true if n to the k power ends with n.

    This code won't work correctly for integral powers of 10, but they can
    never be morphic anyway, except for 1, which I handle specially.
    '''
    if n == 1:
        return 1

    modulo = power( 10, ceil( log10( n ) ) )
    powmod = getPowMod( n, k, modulo )

    return 1 if ( n == powmod ) else 0


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 1 ) ] )
def isKMorphicOperator( n, k ):
    return isKMorphic( n, k )


def isAutomorphic( n ):
    return isKMorphic( n, 2 )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isAutomorphicOperator( n ):
    return isKMorphic( n, 2 )


def isTrimorphic( n ):
    return isKMorphic( n, 3 )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isTrimorphicOperator( n ):
    return isKMorphic( n, 3 )


#******************************************************************************
#
#  getLeftTruncationsOperator
#
#******************************************************************************

def getLeftTruncationsGenerator( n ):
    n = getMPFIntegerAsString( n )

    for i, _ in enumerate( n ):
        yield mpmathify( n[ i : ] )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def getLeftTruncationsOperator( n ):
    return RPNGenerator.createGenerator( getLeftTruncationsGenerator, n )


#******************************************************************************
#
#  getRightTruncationsOperator
#
#******************************************************************************

def getRightTruncationsGenerator( n ):
    result = getMPFIntegerAsString( n )

    for i in range( len( result ), 0, -1 ):
        yield mpmathify( result[ 0 : i ] )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def getRightTruncationsOperator( n ):
    return RPNGenerator.createGenerator( getRightTruncationsGenerator, n )


#******************************************************************************
#
#  getMultiplicativePersistence
#
#******************************************************************************

def getMultiplicativePersistence( n, exponent = 1, dropZeroes = False, persistence = 0 ):
    if exponent == 1 and n < 10:
        return persistence

    if n <= 1:
        return persistence

    return getMultiplicativePersistence( multiplyDigitList( n, exponent, dropZeroes ),
                                         exponent, dropZeroes, persistence + 1 )


def getPersistence( n ):
    return getMultiplicativePersistence( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def getPersistenceOperator( n ):
    return getMultiplicativePersistence( n )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 1 ) ] )
def getKPersistenceOperator( n, k ):
    return getMultiplicativePersistence( n, k )


def getErdosPersistence( n ):
    return getMultiplicativePersistence( n, 1, True )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def getErdosPersistenceOperator( n ):
    return getMultiplicativePersistence( n, 1, True )


#******************************************************************************
#
#  showMultiplicativePersistence
#
#******************************************************************************

def showMultiplicativePersistenceGenerator( n, exponent = 1, dropZeroes = False ):
    yield n

    if exponent == 1:
        while n >= 10:
            n = multiplyDigitList( n, exponent, dropZeroes )
            yield n
    else:
        while n > 1:
            n = multiplyDigitList( n, exponent, dropZeroes )
            yield n


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def showPersistenceOperator( n ):
    return RPNGenerator.createGenerator( showMultiplicativePersistenceGenerator, [ n ] )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 1 ) ] )
def showKPersistenceOperator( n, k ):
    return RPNGenerator.createGenerator( showMultiplicativePersistenceGenerator, [ n, k ] )


#******************************************************************************
#
#  showErdosPersistenceGenerator
#
#******************************************************************************

def showErdosPersistenceGenerator( n ):
    yield n

    while n >= 10:
        n = multiplyDigitList( n, 1, True )
        yield n


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def showErdosPersistenceOperator( n ):
    return RPNGenerator.createGenerator( showErdosPersistenceGenerator, [ n ] )


#******************************************************************************
#
#  permuteDigitsOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def permuteDigitsOperator( n ):
    return RPNGenerator.createPermutations( getMPFIntegerAsString( n ) )


#******************************************************************************
#
#  isIncreasingOperator
#
#******************************************************************************

def isIncreasing( n ):
    n = getMPFIntegerAsString( n )

    for i in range( 1, len( n ) ):
        if n[ i ] < n[ i - 1 ]:
            return 0

    return 1


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isIncreasingOperator( n ):
    return isIncreasing( n )


#******************************************************************************
#
#  isDecreasingOperator
#
#******************************************************************************

def isDecreasing( n ):
    n = getMPFIntegerAsString( n )

    for i in range( 1, len( n ) ):
        if n[ i ] > n[ i - 1 ]:
            return 0

    return 1


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isDecreasingOperator( n ):
    return isDecreasing( n )


#******************************************************************************
#
#  isBouncyOperator
#
#******************************************************************************

def isBouncy( n ):
    if isIncreasing( n ) == 0 and isDecreasing( n ) == 0:
        return 1

    return 0


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isBouncyOperator( n ):
    return isBouncy( n )


#******************************************************************************
#
#  rotateDigitsLeftOperator
#
#******************************************************************************

def rotateDigitsLeft( n, k ):
    if k < 0:
        return rotateDigitsRight( n, fneg( k ) )

    n = getMPFIntegerAsString( n )

    if k > len( n ):
        fmod( k, len( n ) )

    rotate = int( k )

    n = n[ rotate : ] + n[ : rotate ]
    return mpmathify( n )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( ) ] )
def rotateDigitsLeftOperator( n, k ):
    return rotateDigitsLeft( n, k )


#******************************************************************************
#
#  rotateDigitsRightOperator
#
#******************************************************************************

def rotateDigitsRight( n, k ):
    if k < 0:
        return rotateDigitsLeft( n, fneg( k ) )

    n = getMPFIntegerAsString( n )

    if k > len( n ):
        fmod( k, len( n ) )

    rotate = int( k )

    n = n[ -rotate : ] + n[ : -rotate ]
    return mpmathify( n )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( ) ] )
def rotateDigitsRightOperator( n, k ):
    return rotateDigitsRight( n, k )


#******************************************************************************
#
#  getCyclicPermutations
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def getCyclicPermutationsOperator( n ):
    result = [ n ]

    n = getMPFIntegerAsString( n )

    for _ in range( len( n ) - 1 ):
        n = n[ 1 : ] + n[ : 1 ]
        result.append( mpmathify( n ) )

    return result


#******************************************************************************
#
#  isDigitalPermutationOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 0 ) ] )
def isDigitalPermutationOperator( n, k ):
    str1 = getMPFIntegerAsString( n )
    str2 = getMPFIntegerAsString( k )

    if len( str1 ) != len( str2 ):
        return 0

    if sorted( str1 ) != sorted( str2 ):
        return 0

    return 1


#******************************************************************************
#
#  getSquareDigitChainOperator
#
#******************************************************************************

def generateSquareDigitChainGenerator( n ):
    n = int( n )

    chain = [ ]

    if n == 0:
        yield 0
        return

    if n == 1:
        yield 1
        return

    done = False

    while not done:
        digits = getDigitList( n )

        n = 0

        for i in digits:
            n = fadd( n, pow( i, 2 ) )

        if n in chain:
            done = True
        else:
            chain.append( n )

            yield n


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def generateSquareDigitChainOperator( n ):
    return RPNGenerator.createGenerator( generateSquareDigitChainGenerator, [ n ] )


#******************************************************************************
#
#  isStepNumberOperator
#
#******************************************************************************

def isStepNumber( n ):
    if n < 10:
        return 0

    n = getMPFIntegerAsString( n )

    for i in range( 1, len( n ) ):
        if abs( int( n[ i ] ) - int( n[ i - 1 ] ) ) != 1:
            return 0

    return 1


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isStepNumberOperator( n ):
    return isStepNumber( n )


#******************************************************************************
#
#  buildNextStepNumbers
#
#  This method expects a list of digits.
#
#******************************************************************************

def buildNextStepNumbers( n ):
    result = [ ]

    if n[ -1 ] == 0:
        result.append( n + [ 1 ] )
    elif n[ -1 ] == 9:
        result.append( n + [ 8 ] )
    else:
        result.append( n + [ n[ -1 ] - 1 ] )
        result.append( n + [ n[ -1 ] + 1 ] )

    return result


#******************************************************************************
#
#  buildStepNumbersOperator
#
#******************************************************************************

def buildStepNumbersGenerator( maxLength ):
    if maxLength < 2:
        raise ValueError( '\'build_step_numbers\' requires an argument of 2 or greater' )

    stepNumbers = [ ]

    for i in range( 1, 10 ):
        stepNumbers.extend( buildNextStepNumbers( [ i ] ) )

    for i in stepNumbers:
        yield combineDigits( i )

    for i in range( 0, int( maxLength ) - 2 ):
        newStepNumbers = [ ]

        for j in stepNumbers:
            newStepNumbers.extend( buildNextStepNumbers( j ) )

        stepNumbers = newStepNumbers

        for j in stepNumbers:
            yield combineDigits( j )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def buildStepNumbersOperator( n ):
    return RPNGenerator.createGenerator( buildStepNumbersGenerator, [ n ] )


#******************************************************************************
#
#  isSmithNumberOperator
#
#******************************************************************************

@cachedFunction( 'smith' )
def isSmithNumber( n ):
    if isPrime( n ) or n < 2:
        return 0

    sum1 = sumDigits( n )
    sum2 = fsum( [ sumDigits( i ) for i in getFactors( n ) ] )

    return 1 if sum1 == sum2 else 0


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def isSmithNumberOperator( n ):
    isSmithNumber( n )


#******************************************************************************
#
#  isBaseKSmithNumberOperator
#
#******************************************************************************

@cachedFunction( 'base_k_smith' )
def isBaseKSmithNumber( n, k ):
    if k > 4 and n == 4:
        return 1

    if k > 4 and n < k:
        return 0

    if isPrime( n ) or n < 2:
        return 0

    sum1 = fsum( getBaseKDigits( n, k ) )
    sum2 = fsum( [ fsum( getBaseKDigits( i, k ) ) for i in getFactors( n ) ] )

    return 1 if sum1 == sum2 else 0


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 2 ) ] )
def isBaseKSmithNumberOperator( n, k ):
    return isBaseKSmithNumber( n, k )


#******************************************************************************
#
#  isOrderKSmithNumberOperator
#
#******************************************************************************

@cachedFunction( 'order_k_smith' )
def isOrderKSmithNumber( n, k ):
    if isPrime( n ) or n < 2:
        return 0

    digitList1 = getDigitList( n, dropZeroes=True )
    digitList2 = [ item for sublist in [ getDigitList( m, dropZeroes=True ) for m in getFactors( n ) ]
                   for item in sublist ]

    if sorted( digitList1 ) == sorted( digitList2 ):
        return 0

    sum1 = fsum( [ power( i, k ) for i in getDigitList( n ) ] )
    sum2 = fsum( [ power( j, k ) for j in [ item for sublist in
                                            [ getDigitList( m ) for m in getFactors( n ) ] for item in sublist ] ] )

    return 1 if sum1 == sum2 else 0


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 1 ) ] )
def isOrderKSmithNumberOperator( n, k ):
    isOrderKSmithNumber( n, k )


#  Look and Say from:
#  https://www.rosettacode.org/wiki/Look-and-say_sequence#Python

#>>> from itertools import groupby
#>>> def lookandsay(number):
#    return ''.join( str(len(list(g))) + k
#                for k,g in groupby(number) )
#
#>>> numberstring='1'
#>>> for i in range(10):
#    print numberstring
#    numberstring = lookandsay(numberstring)
#



#!/usr/bin/env python

#******************************************************************************
#
#  rpnInput.py
#
#  rpnChilada input functions
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import math
import os

import arrow

from mpmath import fdiv, fneg, mp, mpc, mpf, mpmathify

from rpn.rpnDateTime import RPNDateTime
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnSettings import setAccuracy
from rpn.rpnUtils import oneArgFunctionEvaluator
from rpn.rpnValidator import argValidator, StringValidator

import rpn.rpnGlobals as g


#******************************************************************************
#
#  convertToBase10
#
#******************************************************************************

def convertToBase10( integer, mantissa, inputRadix ):
    result = mpmathify( 0 )
    base = mpmathify( 1 )

    validNumerals = g.numerals[ : inputRadix ]

    for i in range( len( integer ) - 1, -1, -1 ):
        digit = validNumerals.find( integer[ i ] )

        if digit == -1:
            raise ValueError( 'invalid numeral \'%c\' for base %d' % ( integer[ i ], inputRadix ) )

        result += digit * base
        base *= inputRadix

    base = fdiv( 1, inputRadix )

    for i, numeral in enumerate( mantissa ):
        digit = validNumerals.find( numeral )

        if digit == -1:
            raise ValueError( 'invalid numeral \'%c\' for base %d' % ( mantissa[ i ], inputRadix ) )

        result += digit * base
        base /= inputRadix

    return result


#******************************************************************************
#
#  parseInputValue
#
#  Parse out a datetime value or a numerical expression and attempt to set
#  the precision to an appropriate value based on the expression.
#
#******************************************************************************

def parseInputValue( term, inputRadix = 10 ):
    if isinstance( term, mpf ):
        return term

    possibleDate = False

    hasDigit = False
    hasDatePunct = False
    illegalDateChars = False

    # one 'T' is allowed in a date, but no other letters
    for c in term:
        if c.isdigit( ):
            hasDigit = True
            break

    for c in term[ 1 : ]:
        if c in ' -/:':
            hasDatePunct = True
            break

    for c in term:
        if c in '[]':
            illegalDateChars = True
            break

    if hasDigit and hasDatePunct and not illegalDateChars:
        possibleDate = True

    if not g.interactive:
        if term[ 0 ] == '$' and term[ 1 : ] in g.userVariables:
            term = parseInputValue( g.userVariables[ term[ 1 : ] ] )
            return term

        if term[ 0 ] == '@' and term[ 1 : ] in g.userFunctions:
            term = g.userFunctions[ term[ 1 : ] ]
            return term

        if term[ 0 ] == "'":
            return term[ 1 : ]

    innerChars = term[ 1 : -1 ]

    # this helps us parse dates
    if '/' in innerChars and possibleDate:
        term = term.replace( '/', '-' )
        innerChars = term[ 1 : -1 ]

    # Try to parse a date
    if possibleDate:
        tryAgain = False
        datetime = None

        # Try to parse a default date-time
        try:
            datetime = arrow.get( term )

            # convert arrow to RPNDateTime with the parsed timezone
            datetime = RPNDateTime( datetime.year, datetime.month, datetime.day,
                                    datetime.hour, datetime.minute, datetime.second,
                                    datetime.microsecond )
            tryAgain = False
        except arrow.parser.ParserError:
            tryAgain = True

        if tryAgain:
            try:
                datetime = arrow.get( term,
                                      [
                                          'MMM D YYYY HH:mm:ss ZZ',
                                          'MMM D, YYYY HH:mm:ss ZZ',
                                          'MMM DD YYYY HH:mm:ss ZZ',
                                          'MMM DD, YYYY HH:mm:ss ZZ',
                                          'MMMM D YYYY HH:mm:ss ZZ',
                                          'MMMM D, YYYY HH:mm:ss ZZ',
                                          'MMMM DD YYYY HH:mm:ss ZZ',
                                          'MMMM DD, YYYY HH:mm:ss ZZ',
                                          'YYYY-MM-DD HH:mm:ss ZZ',
                                      ] )

                # convert arrow to RPNDateTime with the parsed timezone
                datetime = RPNDateTime( datetime.year, datetime.month, datetime.day,
                                        datetime.hour, datetime.minute, datetime.second,
                                        datetime.microsecond, datetime.tzinfo )
                tryAgain = False
            except arrow.parser.ParserError:
                tryAgain = True

        # If that fails, try to parse without a timezone, and use the local timezone
        if tryAgain:
            try:
                datetime = arrow.get( term,
                                      [
                                          'MMM D YYYY HH:mm:ss',
                                          'MMM D YYYY',
                                          'MMM D, YYYY HH:mm:ss',
                                          'MMM D, YYYY',
                                          'MMM DD YYYY HH:mm:ss',
                                          'MMM DD YYYY',
                                          'MMM DD, YYYY HH:mm:ss',
                                          'MMM DD, YYYY',
                                          'MMMM D YYYY HH:mm:ss',
                                          'MMMM D YYYY',
                                          'MMMM D, YYYY HH:mm:ss',
                                          'MMMM D, YYYY',
                                          'MMMM DD YYYY HH:mm:ss',
                                          'MMMM DD YYYY',
                                          'MMMM DD, YYYY HH:mm:ss',
                                          'MMMM DD, YYYY',
                                          'YYYY-MM-DD HH:mm:ss',
                                          'YYYY-MM-DD',
                                          'MM-DD-YYYY',
                                      ] )

                # convert arrow to RPNDateTime with the local (default) timezone
                datetime = RPNDateTime( datetime.year, datetime.month, datetime.day,
                                        datetime.hour, datetime.minute, datetime.second,
                                        datetime.microsecond )
            except arrow.parser.ParserError:
                # We couldn't parse a date, but it might be something else (like a dice expression)
                pass

        # if we get a datetime, then let's use it, otherwise keep trying
        if datetime:
            return datetime

    if term == '0':
        return mpmathify( 0 )

    # ignore commas
    term = ''.join( [ i for i in term if i not in ',' ] )

    if term[ 0 ] == '\\':
        term = term[ 1 : ]
        ignoreSpecial = True
    else:
        ignoreSpecial = False

    if inputRadix == 10:
        imaginary = term[ -1 ] in ( 'i', 'j' )
    else:
        imaginary = False

    if imaginary:
        term = term[ : -1 ]

    if '.' in term:
        if inputRadix == 10:
            newPrecision = len( term ) + 1

            if mp.dps < newPrecision:
                setAccuracy( newPrecision )

            if imaginary:
                return mpc( imag = term )

            return mpmathify( term )

        decimal = term.find( '.' )
    else:
        decimal = len( term )

    negative = term[ 0 ] == '-'

    if negative:
        start = 1
    else:
        if term[ 0 ] == '+':
            start = 1
        else:
            start = 0

    integer = term[ start : decimal ]
    mantissa = term[ decimal + 1 : ]

    # check for hex, then binary, then octal, otherwise a plain old decimal integer
    if not ignoreSpecial and mantissa == '':
        if integer[ 0 ] == '0':
            if len( integer ) == 1:
                return mpmathify( 0 )

            if integer[ 1 ] in 'Xx':
                # set the precision big enough to handle this value
                newPrecision = math.ceil( ( math.log10( 16 ) * ( len( integer ) - 2 ) ) ) + 1

                if mp.dps < newPrecision:
                    setAccuracy( newPrecision )

                return mpmathify( int( integer[ 2 : ], 16 ) )

            if integer[ -1 ] in 'bB':
                # set the precision big enough to handle this value
                newPrecision = math.ceil( math.log10( 2 ) * ( len( integer ) - 1 ) ) + 1

                if mp.dps < newPrecision:
                    setAccuracy( newPrecision )

                integer = integer[ : -1 ]
                return mpmathify( int( integer, 2 ) * ( -1 if negative else 1 ) )

            integer = integer[ 1 : ]

            return mpmathify( int( integer, 8 ) )
        if integer[ 0 ] == '1' and integer[ -1 ] in 'bB':
            # set the precision big enough to handle this value
            newPrecision = math.ceil( math.log10( 2 ) * ( len( integer ) - 1 ) ) + 1

            if mp.dps < newPrecision:
                setAccuracy( newPrecision )

            integer = integer[ : -1 ]
            return mpmathify( int( integer, 2 ) * ( -1 if negative else 1 ) )

        if inputRadix == 10:
            newPrecision = len( integer ) + 1

            if mp.dps < newPrecision:
                setAccuracy( newPrecision )

            result = fneg( integer ) if negative else mpmathify( integer )

            if imaginary:
                return mpc( real = '0.0', imag = result )

            return result

    # finally, we have a non-radix 10 number with a mantissa to parse
    result = convertToBase10( integer, mantissa, inputRadix )
    result = fneg( result ) if negative else mpmathify( result )

    if imaginary:
        return mpc( real = '0.0', imag = result )

    return result


#******************************************************************************
#
#  readListFromFile
#
#******************************************************************************

def readListFromFileGenerator( filename ):
    filename = g.cwd + os.sep + filename

    with open( filename ) as file:
        for i in file:
            if i == '\n':
                continue

            i = i[ : -1 ]    # trim off the newline

            yield parseInputValue( i )


@oneArgFunctionEvaluator( )
@argValidator( [ StringValidator( ) ] )
def readListFromFileOperator( filename ):
    return RPNGenerator( readListFromFileGenerator( filename ) )


#******************************************************************************
#
#  readNUmberFromFile
#
#******************************************************************************

def readNumberFromFileGenerator( filename ):
    filename = g.cwd + os.sep + filename

    string = ''

    with open( filename ) as file:
        for i in file:
            for c in i:
                if '0' <= c <= '9':
                    string += c

    setAccuracy( len( string ) + 1 )

    return mpmathify( string )


@oneArgFunctionEvaluator( )
@argValidator( [ StringValidator( ) ] )
def readNumberFromFileOperator( filename ):
    return readNumberFromFileGenerator( filename )



#!/usr/bin/env python

#******************************************************************************
#
#  rpnList.py
#
#  rpnChilada list operators
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import collections
import itertools
import random

from mpmath import arange, fadd, fdiv, fmod, fmul, fneg, fprod, fsub, fsum, inf, power, root, sqrt

from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMath import add, multiply, square, subtract, divide
from rpn.rpnMeasurementClass import RPNMeasurement
from rpn.rpnUtils import getPowerSet, listArgFunctionEvaluator, listAndOneArgFunctionEvaluator, \
                         listAndTwoArgFunctionEvaluator, twoArgFunctionEvaluator
from rpn.rpnValidator import argValidator, ComparableValidator, ComplexValidator, DefaultValidator, IntValidator, \
                             ListValidator


#******************************************************************************
#
#  getGCD
#
#******************************************************************************

def getGCD( n, k ):
    while k:
        n, k = k, fmod( n, k )

    return n


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ), IntValidator( 0 ) ] )
def getGCDOperator( n, k ):
    return getGCD( n, k )


#******************************************************************************
#
#  getGCDOfListOperator
#
#******************************************************************************

def getGCDOfList( args ):
    if isinstance( args, RPNGenerator ):
        args = list( args )

    if not isinstance( args, list ):
        args = [ args ]

    if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
        return [ getGCDOfList( arg ) for arg in args ]

    result = set( )

    if len( args ) == 1:
        return args[ 0 ]

    for pair in itertools.combinations( args, 2 ):
        result.add( getGCD( *pair ) )

    if len( result ) == 1:
        return result.pop( )

    return getGCDOfList( list( result ) )


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getGCDOfListOperator( n ):
    return getGCDOfList( n )


#******************************************************************************
#
#  alternateSigns
#
#******************************************************************************

def alternateSigns( n, startNegative = False ):
    negative = startNegative

    for i in n:
        yield fneg( i ) if negative else i
        negative = not negative


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def alternateSignsOperator( n ):
    return RPNGenerator( alternateSigns( n, False ) )


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def alternateSigns2Operator( n ):
    return RPNGenerator( alternateSigns( n, True ) )


#******************************************************************************
#
#  getAlternatingSum
#
#******************************************************************************

def getAlternatingSum( args, startNegative = False ):
    result = 0

    negative = startNegative

    for i in args:
        if negative:
            result = fsub( result, i )
        else:
            result = fadd( result, i )

        negative = not negative

    return result


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getAlternatingSumOperator( n ):
    return getAlternatingSum( n, False )


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getAlternatingSum2Operator( n ):
    return getAlternatingSum( n, True )


#******************************************************************************
#
#  appendLists
#
#******************************************************************************

def appendLists( arg1, arg2 ):
    result = list( arg1 )
    result.extend( list( arg2 ) )

    return result


@argValidator( [ ListValidator( ), ListValidator( ) ] )
def appendListsOperator( n, k ):
    return appendLists( n, k )


#******************************************************************************
#
#  compareListsOperator
#
#******************************************************************************

@argValidator( [ ListValidator( ), ListValidator( ) ] )
def compareListsOperator( arg1, arg2 ):
    if len( arg1 ) != len( arg2 ):
        return 0

    for i, arg in enumerate( arg1 ):
        if arg != arg2[ i ]:
            return 0

    return 1


#******************************************************************************
#
#  countElementsOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def countElementsOperator( args ):
    if isinstance( args, list ):
        return len( args )

    if isinstance( args, RPNGenerator ) and args.getCount( ) > -1:
        return args.getCount( )

    # args might be a generator, so we need to iterate
    count = 0

    for _ in args:
        count += 1

    return count


#******************************************************************************
#
#  interleaveOperator
#
#  http://stackoverflow.com/questions/7946798/interleaving-two-lists-in-python
#
#******************************************************************************

@argValidator( [ ListValidator( ), ListValidator( ) ] )
def interleaveOperator( arg1, arg2 ):
    return [ val for pair in zip( arg1, arg2 ) for val in pair ]


#******************************************************************************
#
#  collateOperator
#
#******************************************************************************

def collate( argList ):
    if not isinstance( argList, list ):
        yield argList
        return

    listOfLists = [ ]
    length = 0

    for arg in argList:
        if isinstance( arg, list ):
            listOfLists.append( arg )

            if length < len( arg ):
                length = len( arg )
        else:
            listOfLists.append( [ arg ] )

            if length < 1:
                length = 1

    for i in range( 0, length ):
        newSubList = [ ]

        for subList in argList:
            newSubList.append( subList[ i ] )

        yield newSubList


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def collateOperator( n ):
    return RPNGenerator( collate( n ) )


#******************************************************************************
#
#  makeUnionOperator
#
#******************************************************************************

@argValidator( [ ListValidator( ), ListValidator( ) ] )
def makeUnionOperator( arg1, arg2 ):
    result = set( arg1 )
    return list( result.union( set( arg2 ) ) )


#******************************************************************************
#
#  makeIntersectionOperator
#
#******************************************************************************

@argValidator( [ ListValidator( ), ListValidator( ) ] )
def makeIntersectionOperator( arg1, arg2 ):
    result = set( arg1 )
    return list( result.intersection( set( arg2 ) ) )


#******************************************************************************
#
#  getDifferenceOperator
#
#******************************************************************************

@argValidator( [ ListValidator( ), ListValidator( ) ] )
def getDifferenceOperator( arg1, arg2 ):
    return list( set( arg2 ) - set( arg1 ) )


#******************************************************************************
#
#  getIndexOfMaxOperator
#
#******************************************************************************

#@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getIndexOfMaxOperator( args ):
    maximum = -inf
    result = -1
    index = 0

    for i in args:
        if i > maximum:
            maximum = i
            result = index

        index += 1

    return result


#******************************************************************************
#
#  getIndexOfMinOperator
#
#******************************************************************************

#@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getIndexOfMinOperator( args ):
    minimum = inf
    result = -1
    index = 0

    for i in args:
        if i < minimum:
            minimum = i
            result = index

        index += 1

    return result


#******************************************************************************
#
#  getListElementOperator
#
#******************************************************************************

@argValidator( [ ListValidator( ), IntValidator( ) ] )
def getListElementOperator( args, index ):
    if isinstance( index, ( list, RPNGenerator ) ):
        result = [ ]

        for i in index:
            result.append( args[ int( i ) ] )

        return result

    return args[ int( index ) ]


#******************************************************************************
#
#  getRandomElementOperator
#
#******************************************************************************

@argValidator( [ ListValidator( ) ] )
def getRandomElementOperator( arg ):
    return random.choice( arg )


#******************************************************************************
#
#  enumerateListGenerator
#
#******************************************************************************

def enumerateListGenerator( args, k ):
    i = 0

    for arg in args:
        yield [ fadd( i, k ), arg ]
        i += 1


@argValidator( [ ListValidator( ), IntValidator( ) ] )
def enumerateListOperator( args, k ):
    return RPNGenerator.createGenerator( enumerateListGenerator, [ args, k ] )


#******************************************************************************
#
#  getSliceGenerator
#
#******************************************************************************

def getSliceGenerator( args, start, end ):
    if end == 0:
        for i in args[ int( start ) : ]:
            yield i
    else:
        for i in args[ int( start ) : int( end ) ]:
            yield i


@listAndTwoArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), IntValidator( ), IntValidator( ) ] )
def getSliceOperator( args, start, end ):
    return RPNGenerator.createGenerator( getSliceGenerator, [ args, start, end ] )


#******************************************************************************
#
#  getSublistGenerator
#
#******************************************************************************

def getSublistGenerator( args, start, count ):
    for i in arange( start, fadd( start, count ) ):
        yield args[ int( i ) ]


@listAndTwoArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), IntValidator( ), IntValidator( ) ] )
def getSublistOperator( args, start, count ):
    return RPNGenerator.createGenerator( getSublistGenerator, [ args, start, count ] )


#******************************************************************************
#
#  getLeftOperator
#
#******************************************************************************

@listAndOneArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), IntValidator( ) ] )
def getLeftOperator( args, count ):
    result = [ ]

    if isinstance( count, list ):
        for i in count:
            result.append( args[ : int( i ) ] )
    else:
        result.append( args[ : int( count ) ] )

    if len( result ) == 1:
        return result[ 0 ]

    return result


#******************************************************************************
#
#  getRightOperator
#
#******************************************************************************

@listAndOneArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), IntValidator( ) ] )
def getRightOperator( args, count ):
    result = [ ]

    if isinstance( count, list ):
        for i in count:
            result.append( args[ int( fneg( i ) ) : ] )
    else:
        result.append( args[ int( fneg( count ) ) : ] )

    if len( result ) == 1:
        return result[ 0 ]

    return result


#******************************************************************************
#
#  getListDiffsOperator
#
#******************************************************************************

def getListDiffs( args ):
    old = None

    for i in args:
        if old is not None:
            yield subtract( i, old )

        old = i


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getListDiffsOperator( n ):
    return RPNGenerator( getListDiffs( n ) )


#******************************************************************************
#
#  getCumulativeListDiffsOperator
#
#******************************************************************************

def getCumulativeListDiffs( args ):
    first = None

    for i in args:
        if first is None:
            first = i
        else:
            yield subtract( i, first )


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getCumulativeListDiffsOperator( n ):
    return RPNGenerator( getCumulativeListDiffs( n ) )


#******************************************************************************
#
#  getCumulativeListProductsOperator
#
#******************************************************************************

def getCumulativeListProducts( args ):
    total = 1

    for i in args:
        total = multiply( total, i )
        yield total


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getCumulativeListProductsOperator( n ):
    return RPNGenerator( getCumulativeListProducts( n ) )


#******************************************************************************
#
#  getCumulativeListMeansOperator
#
#******************************************************************************

def getCumulativeListMeans( args ):
    mean = None

    for index, i in enumerate( args ):
        if mean is None:
            mean = i
        else:
            mean = fdiv( fadd( fmul( mean, index ), i ), fadd( index, 1 ) )

        yield mean


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getCumulativeListMeansOperator( n ):
    return RPNGenerator( getCumulativeListMeans( n ) )


#******************************************************************************
#
#  getCumulativeListSumsOperator
#
#******************************************************************************

def getCumulativeListSums( args ):
    total = 0

    for i in args:
        total = add( total, i )
        yield total


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getCumulativeListSumsOperator( n ):
    return RPNGenerator( getCumulativeListSums( n ) )


#******************************************************************************
#
#  getListRatiosOperator
#
#******************************************************************************

def getListRatios( args ):
    old = None

    for i in args:
        if old is not None:
            yield divide( i, old )

        old = i


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getListRatiosOperator( n ):
    return RPNGenerator( getListRatios( n ) )


#******************************************************************************
#
#  getCumulativeListRatiosOperator
#
#******************************************************************************

def getCumulativeListRatios( args ):
    first = None

    for i in args:
        if first is None:
            first = i
        else:
            yield divide( i, first )


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getCumulativeListRatiosOperator( n ):
    return RPNGenerator( getCumulativeListRatios( n ) )


#******************************************************************************
#
#  getReverseOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getReverseOperator( args ):
    # These list comprehensions _are_ needed.
    # pylint: disable=unnecessary-comprehension
    return [ i for i in reversed( [ j for j in args ] ) ]


#******************************************************************************
#
#  shuffleListOperator
#
#******************************************************************************

def shuffleList( args ):
    if isinstance( args, RPNGenerator ):
        return shuffleList( list( args ) )

    if not isinstance( args, list ):
        return args

    if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
        result = [ ]

        for arg in args:
            result.append( shuffleList( arg ) )

        return result

    result = args[ : ]
    random.shuffle( result )
    return result


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def shuffleListOperator( n ):
    return shuffleList( n )


#******************************************************************************
#
#  getUniqueElementsOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getUniqueElementsOperator( args ):
    return list( set( args ) )


#******************************************************************************
#
#  sortAscendingOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def sortAscendingOperator( args ):
    result = [ ]

    if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
        for arg in args:
            result.append( sorted( arg ) )

        return result

    return sorted( args )


#******************************************************************************
#
#  sortDescendingOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def sortDescendingOperator( args ):
    result = [ ]

    if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
        for arg in args:
            result.append( sorted( arg, reverse = True ) )

        return result

    return sorted( args, reverse = True )


#******************************************************************************
#
#  calculatePowerTowerOperator
#
#******************************************************************************

def calculatePowerTower( args ):
    if isinstance( args, RPNGenerator ):
        return calculatePowerTower( list( args ) )

    if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
        return [ calculatePowerTower( arg ) for arg in args ]

    result = args[ 0 ]

    for i in args[ 1 : ]:
        result = power( result, i )

    return result


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def calculatePowerTowerOperator( n ):
    return calculatePowerTower( n )


#******************************************************************************
#
#  calculatePowerTowerRightOperator
#
#******************************************************************************

def calculatePowerTowerRight( args ):
    if isinstance( args, RPNGenerator ):
        return calculatePowerTowerRight( list( args ) )

    if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
        return [ calculatePowerTowerRight( arg ) for arg in args ]

    result = args[ -1 ]

    for i in args[ -2 : : -1 ]:
        result = power( i, result )

    return result


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def calculatePowerTowerRightOperator( n ):
    return calculatePowerTowerRight( n )


#******************************************************************************
#
#  getSumOperator
#
#******************************************************************************

def getSum( n ):
    if isinstance( n, RPNGenerator ):
        return getSum( list( n ) )

    if isinstance( n[ 0 ], ( list, RPNGenerator ) ):
        return [ getSum( arg ) for arg in n ]

    result = None

    try:
        result = fsum( n )
    except TypeError:
        result = n[ 0 ]

        for i in n[ 1 : ]:
            result = add( result, i )

    return result


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getSumOperator( n ):
    return getSum( n )


#******************************************************************************
#
#  getProductOperator
#
#******************************************************************************

def getProduct( n ):
    if isinstance( n, RPNGenerator ):
        return getProduct( list( n ) )

    if isinstance( n[ 0 ], ( list, RPNGenerator ) ):
        return [ getProduct( arg ) for arg in n ]

    if not n:
        return 0

    if len( n ) == 1:
        return n[ 0 ]

    hasUnits = False

    for item in n:
        if isinstance( item, RPNMeasurement ):
            hasUnits = True
            break

    if hasUnits:
        result = RPNMeasurement( 1 )

        for item in n:
            if isinstance( item, list ):
                return [ getProduct( arg ) for arg in item ]

            result = multiply( result, item )

        return result

    if not n:
        return 0

    if isinstance( n[ 0 ], list ):
        return [ getProduct( item ) for item in n ]

    return fprod( n )


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getProductOperator( n ):
    return getProduct( n )


#******************************************************************************
#
#  getStandardDeviationOperator
#
#******************************************************************************

def getStandardDeviation( args ):
    if isinstance( args, RPNGenerator ):
        return getStandardDeviation( list( args ) )

    if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
        return [ getStandardDeviation( arg ) for arg in args ]

    if len( args ) < 2:
        return 0

    mean = fsum( args ) / len( args )

    dev = [ power( fsub( i, mean ), 2 ) for i in args ]
    return sqrt( fdiv( fsum( dev ), fsub( len( dev ), 1 ) ) )


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getStandardDeviationOperator( n ):
    return getStandardDeviation( n )


#******************************************************************************
#
#  reduceListOperator
#
#  This function reduces out the greatest common denominator from a list of
#  values.
#
#******************************************************************************

def reduceList( args ):
    if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
        return [ reduceList( arg ) for arg in args ]

    gcd = getGCDOfList( args )

    result = [ ]

    for i in args:
        result.append( fdiv( i, gcd ) )

    return result


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def reduceListOperator( n ):
    return reduceList( n )


#******************************************************************************
#
#  calculateGeometricMeanOperator
#
#******************************************************************************

def calculateGeometricMean( args ):
    if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
        return [ calculateGeometricMean( list( arg ) ) for arg in args ]

    return root( fprod( args ), len( args ) )


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def calculateGeometricMeanOperator( n ):
    return calculateGeometricMean( n )


#******************************************************************************
#
#  calculateHarmonicMeanOperator
#
#******************************************************************************

def calculateHarmonicMean( args ):
    if isinstance( args, RPNGenerator ):
        return calculateHarmonicMean( list( args ) )

    if isinstance( args, list ):
        if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
            return [ calculateHarmonicMean( list( arg ) ) for arg in args ]

        result = 0

        for arg in args:
            result = fadd( result, fdiv( 1, arg ) )

        return fdiv( len( args ), result )

    return args


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def calculateHarmonicMeanOperator( n ):
    return calculateHarmonicMean( n )


#******************************************************************************
#
#  calculateAntiharmonicMeanOperator
#
# https://en.wikipedia.org/wiki/Contraharmonic_mean
#
#******************************************************************************

def calculateAntiharmonicMean( args ):
    if isinstance( args, RPNGenerator ):
        return calculateAntiharmonicMean( list( args ) )

    if isinstance( args, list ):
        if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
            return [ calculateAntiharmonicMean( list( arg ) ) for arg in args ]

        return fdiv( fsum( args, squared=True ), fsum( args ) )

    return args


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def calculateAntiharmonicMeanOperator( n ):
    return calculateAntiharmonicMean( n )


#******************************************************************************
#
#  calculateArithmeticMeanOperator
#
#******************************************************************************

def calculateArithmeticMean( args ):
    if isinstance( args, RPNGenerator ):
        total = 0
        count = 0

        for i in args:
            total += i
            count += 1

        return fdiv( total, count )

    if isinstance( args, list ):
        if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
            return [ calculateArithmeticMean( list( arg ) ) for arg in args ]

        if isinstance( args[ 0 ], RPNMeasurement ):
            # TODO: handle measurements
            raise ValueError( '\'mean\' doesn\'t support measurements (yet)' )

        return fdiv( fsum( args ), len( args ) )

    return args


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def calculateArithmeticMeanOperator( n ):
    return calculateArithmeticMean( n )


#******************************************************************************
#
#  calculateRootMeanSquareOperator
#
#******************************************************************************

def calculateRootMeanSquare( args ):
    if isinstance( args, RPNGenerator ):
        total = 0
        count = 0

        for i in args:
            total += power( i, 2 )
            count += 1

        return square( fdiv( total, count ) )

    if isinstance( args, list ):
        if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
            return [ calculateRootMeanSquare( list( arg ) ) for arg in args ]

        if isinstance( args[ 0 ], RPNMeasurement ):
            raise ValueError( '\'root_mean_square\' doesn\'t support measurements' )

        return sqrt( fdiv( fsum( args, squared=True ), len( args ) ) )

    return args


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def calculateRootMeanSquareOperator( n ):
    return calculateRootMeanSquare( n )


#******************************************************************************
#
#  getZeroesOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getZeroesOperator( args ):
    if isinstance( args, ( RPNGenerator, list ) ):
        return [ index for index, e in enumerate( args ) if e == 0 ]

    return [ 0 ] if args == 0 else [ ]


#******************************************************************************
#
#  getNonzeroesOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getNonzeroesOperator( args ):
    if isinstance( args, ( RPNGenerator, list ) ):
        return [ index for index, e in enumerate( args ) if e != 0 ]

    return [ 0 ] if args == 0 else [ ]


#******************************************************************************
#
#  groupElementsOperator
#
#******************************************************************************

def groupElements( args, count ):
    if isinstance( count, list ):
        return [ groupElements( args, i ) for i in count ]

    if isinstance( args, list ):
        if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
            return [ groupElements( args, count ) for arg in args ]

        result = [ ]

        n = int( count )

        for i in range( 0, len( args ), n ):
            result.append( args[ i : i + n ] )

        return result

    return [ args ]


@listAndOneArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), IntValidator( 1 ) ] )
def groupElementsOperator( args, count ):
    return groupElements( args, count )


#******************************************************************************
#
#  getOccurrencesOperator
#
#******************************************************************************

def getOccurrences( args ):
    if isinstance( args, list ):
        if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
            return [ getOccurrences( arg ) for arg in args ]

        counter = collections.Counter( )

        for i in args:
            counter[ i ] += 1

        result = [ ]

        for i in counter:
            result.append( [ i, counter[ i ] ] )

        return sorted( result )

    return [ [ args, 1 ] ]


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getOccurrencesOperator( n ):
    return getOccurrences( n )


#******************************************************************************
#
#  getOccurrenceRatiosOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getOccurrenceRatiosOperator( args ):
    if isinstance( args, list ):
        count = len( args )

        if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
            return [ getOccurrences( arg ) for arg in args ]

        counter = collections.Counter( )

        for i in args:
            counter[ i ] += 1

        result = [ ]

        for i in counter:
            result.append( [ i, counter[ i ] / count ] )

        return sorted( result )

    return [ [ args, 1 ] ]


#******************************************************************************
#
#  getCumulativeOccurrenceRatiosOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getCumulativeOccurrenceRatiosOperator( args ):
    if isinstance( args, list ):
        count = len( args )

        if isinstance( args[ 0 ], ( list, RPNGenerator ) ):
            return [ getOccurrences( arg ) for arg in args ]

        counter = collections.Counter( )

        for i in args:
            counter[ i ] += 1

        result = [ ]

        runningTotal = 0
        for i in counter:
            runningTotal += counter[ i ]
            result.append( [ i, runningTotal / count ] )

        return sorted( result )

    return [ [ args, 1 ] ]


#******************************************************************************
#
#  flattenOperator
#
#******************************************************************************

def flatten( value ):
    if isinstance( value, ( list, RPNGenerator ) ):
        result = [ ]

        for item in value:
            result.extend( flatten( item ) )

        return result

    return [ value ]


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def flattenOperator( value ):
    return flatten( value )


#******************************************************************************
#
#  getAndAllOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getAndAllOperator( value ):
    for i in value:
        if not i:
            return 0

    return 1


#******************************************************************************
#
#  getNandAllOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getNandAllOperator( value ):
    for i in value:
        if i:
            return 0

    return 1


#******************************************************************************
#
#  getOrAllOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getOrAllOperator( value ):
    for i in value:
        if i:
            return 1

    return 0


#******************************************************************************
#
#  getNorAllOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getNorAllOperator( value ):
    for i in value:
        if not i:
            return 1

    return 0


#******************************************************************************
#
#  permuteListsOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def permuteListsOperator( lists ):
    for i in lists:
        if not isinstance( i, ( list, RPNGenerator ) ):
            raise ValueError( '\'permute_lists\' expects a list of lists' )

    return RPNGenerator.createProduct( lists )


#******************************************************************************
#
#  getListCombinationsOperator
#
#******************************************************************************

def getListCombinationsGenerator( n, k ):
    for i in itertools.combinations( n, int( k ) ):
        yield list( i )


@listAndOneArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), IntValidator( 1 ) ] )
def getListCombinationsOperator( n, k ):
    if not isinstance( n, ( list, RPNGenerator ) ):
        raise ValueError( '\'get_combinations\' expects a list' )

    if len( n ) < k:
        raise ValueError( 'k must be greater than or equal to the length of list n' )

    return RPNGenerator( getListCombinationsGenerator( n, k ) )


#******************************************************************************
#
#  getListCombinationsWithRepeatsOperator
#
#******************************************************************************

def getListCombinationsWithRepeatsGenerator( n, k ):
    for i in itertools.combinations_with_replacement( n, int( k ) ):
        yield list( i )


@listAndOneArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), IntValidator( 1 ) ] )
def getListCombinationsWithRepeatsOperator( n, k ):
    if not isinstance( n, ( list, RPNGenerator ) ):
        raise ValueError( '\'get_repeat_combinations\' expects a list' )

    if len( n ) < k:
        raise ValueError( 'k must be greater than or equal to the length of list n' )

    return RPNGenerator( getListCombinationsWithRepeatsGenerator( n, k ) )


#******************************************************************************
#
#  getListPermutationsOperator
#
#******************************************************************************

def getListPermutationsGenerator( n, k ):
    for i in itertools.permutations( n, int( k ) ):
        yield list( i )


@listAndOneArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), IntValidator( 1 ) ] )
def getListPermutationsOperator( n, k ):
    if not isinstance( n, ( list, RPNGenerator ) ):
        raise ValueError( '\'permute_list\' expects a list' )

    if len( n ) < k:
        raise ValueError( 'k must be greater than or equal to the length of list n' )

    return RPNGenerator( getListPermutationsGenerator( n, k ) )


#******************************************************************************
#
#  getListPermutationsWithRepeatsOperator
#
#******************************************************************************

def getListPermutationsWithRepeatsGenerator( n, k ):
    for i in itertools.product( n, repeat=int( k ) ):
        yield list( i )


@listAndOneArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), IntValidator( 1 ) ] )
def getListPermutationsWithRepeatsOperator( n, k ):
    if not isinstance( n, ( list, RPNGenerator ) ):
        raise ValueError( '\'get_repeat_permutations\' expects a list' )

    if len( n ) < k:
        raise ValueError( 'k must be greater than or equal to the length of list n' )

    return RPNGenerator( getListPermutationsWithRepeatsGenerator( n, k ) )


#******************************************************************************
#
#  equalsOneOfOperator
#
#******************************************************************************

@argValidator( [ ComplexValidator( ), ListValidator( ) ] )
def equalsOneOfOperator( value, targetList ):
    for i in targetList:
        if value == i:
            return 1

    return 0


#******************************************************************************
#
#  getListPowerSetOperator
#
#******************************************************************************

def getListPowerSet( n ):
    for i in getPowerSet( n ):
        if len( i ) > 0:
            yield list( i )


@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getListPowerSetOperator( n ):
    if isinstance( n, RPNGenerator ):
        return RPNGenerator( getListPowerSet( list( n ) ) )

    return RPNGenerator( getListPowerSet( n ) )


#******************************************************************************

#  findInListOperator
#
#******************************************************************************

@listAndOneArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), DefaultValidator( ) ] )
def findInListOperator( target, k ):
    try:
        result = target.index( k )
    except ValueError:
        return -1

    return result


#******************************************************************************
#
#  isPalindromeListOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def isPalindromeListOperator( n ):
    length = len( n )

    for i in range( 0, length // 2 ):
        if n[ i ] != n[ -( i + 1 ) ]:
            return 0

    return 1


#******************************************************************************
#
#  filterOnFlagsGenerator
#
#******************************************************************************

def filterOnFlagsGenerator( n, k ):
    for nItem, kItem in zip( n, k ):
        if kItem:
            yield nItem


@argValidator( [ ListValidator( ), ListValidator( ) ] )
def filterOnFlagsOperator( n, k ):
    return RPNGenerator.createGenerator( filterOnFlagsGenerator, [ n, k ] )


#******************************************************************************
#
#  filterMaxGenerator
#
#******************************************************************************

def filterMaxGenerator( n, k ):
    for item in n:
        if item <= k:
            yield item


@listAndOneArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), ComparableValidator( ) ] )
def filterMaxOperator( n, k ):
    return RPNGenerator.createGenerator( filterMaxGenerator, [ n, k ] )


#******************************************************************************
#
#  filterMinGenerator
#
#******************************************************************************

def filterMinGenerator( n, k ):
    for item in n:
        if item >= k:
            yield item


@listAndOneArgFunctionEvaluator( )
@argValidator( [ ListValidator( ), ComparableValidator( ) ] )
def filterMinOperator( n, k ):
    return RPNGenerator.createGenerator( filterMinGenerator, [ n, k ] )


#******************************************************************************
#
#  doesListRepeatOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def doesListRepeatOperator( n ):
    n = list( n )
    length = len( n )

    if length == 0:
        return 0

    for i in range( 1, ( length // 2 ) + 1 ):
        substring = n[ 0 : i ]

        start = i
        end = start + i

        foundRepeats = True

        while end <= length:
            if n[ start : end ] != substring:
                foundRepeats = False
                break

            start += i
            end += i

        if not foundRepeats:
            continue

        # Check the remainder, if there is one.
        if ( start < length ) and \
           ( n[ end - i : length ] != substring[ : length - ( end - i ) ] ):
            foundRepeats = False

        if foundRepeats:
            return i

    return 0



#!/usr/bin/env python

#******************************************************************************
#
#  rpnOutput.py
#
#  rpnChilada output functions
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import math
import string
import sys
import textwrap

from mpmath import e, floor, frac, im, mp, mpf, mpmathify, nstr, phi, pi, re, sqrt

from rpn.rpnBase import convertFractionToBaseN, convertToBaseN
from rpn.rpnDateTime import RPNDateTime
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurementClass import RPNMeasurement
from rpn.rpnPersistence import loadHelpData, loadUnitData
from rpn.rpnSpecialBase import convertToFibBase, convertToNonintegerBase, convertToSpecialBase, specialBaseFunctions
from rpn.rpnUnitTypes import basicUnitTypes
from rpn.rpnUtils import addAliases
from rpn.rpnVersion import COPYRIGHT_MESSAGE, PROGRAM_VERSION, PROGRAM_VERSION_STRING, \
                           RPN_PROGRAM_NAME, PROGRAM_DESCRIPTION

import rpn.rpnGlobals as g


#******************************************************************************
#
#  formatNumber
#
#  This takes a mpf value and turns it into a string.
#
#******************************************************************************

def formatNumber( number, outputRadix, leadingZero, integerGrouping, integerDelimiter, decimalDelimiter ):
    negative = ( number < 0 )

    if outputRadix == g.fibBase:
        strInteger = convertToFibBase( floor( number ) )
        strMantissa = ''
    elif outputRadix == g.phiBase:
        strInteger, strMantissa = convertToNonintegerBase( number, phi )
    elif outputRadix == g.eBase:
        strInteger, strMantissa = convertToNonintegerBase( number, e )
    elif outputRadix == g.piBase:
        strInteger, strMantissa = convertToNonintegerBase( number, pi )
    elif outputRadix == g.sqrt2Base:
        strInteger, strMantissa = convertToNonintegerBase( number, sqrt( 2 ) )
    elif outputRadix < 0:   # these mean special bases, not negative numbers
        strInteger = convertToSpecialBase( floor( number ), specialBaseFunctions[ outputRadix ] )
        strMantissa = ''
    elif ( outputRadix != 10 ) or ( g.numerals != g.defaultNumerals ):
        strInteger = str( convertToBaseN( floor( number ), outputRadix, False, g.numerals ) )
        strMantissa = str( convertFractionToBaseN( frac( number ), outputRadix,
                                                   int( mp.dps / math.log10( outputRadix ) ), False ) )
        if strMantissa == '[]':
            strMantissa = ''
    else:
        strNumber = nstr( number, n=g.outputAccuracy, min_fixed=-g.maximumFixed - 1 )

        if '.' in strNumber:
            decimal = strNumber.find( '.' )
        else:
            decimal = len( strNumber )

        strInteger = strNumber[ 1 if negative else 0 : decimal ]

        strMantissa = strNumber[ decimal + 1 : ]

        if strMantissa == '0':
            strMantissa = ''
        elif ( strMantissa != '' ) and ( g.outputAccuracy == -1 ):
            strMantissa = strMantissa.rstrip( '0' )

    if integerGrouping > 0:
        if strInteger[ 0 ] == '-':
            strInteger = strInteger[ 1 : ]

        firstDelimiter = len( strInteger ) % integerGrouping

        if leadingZero and firstDelimiter > 0:
            integerResult = '0' * ( integerGrouping - firstDelimiter )
        else:
            integerResult = ''

        integerResult += strInteger[ : firstDelimiter ]

        for i in range( firstDelimiter, len( strInteger ), integerGrouping ):
            if integerResult != '':
                integerResult += integerDelimiter

            integerResult += strInteger[ i : i + integerGrouping ]
    else:
        integerResult = strInteger

    if g.decimalGrouping > 0:
        mantissaResult = ''

        for i in range( 0, len( strMantissa ), g.decimalGrouping ):
            if mantissaResult != '':
                mantissaResult += decimalDelimiter

            mantissaResult += strMantissa[ i : i + g.decimalGrouping ]
    else:
        mantissaResult = strMantissa

    result = integerResult

    if mantissaResult != '':
        result += '.' + mantissaResult

    return result, negative


#******************************************************************************
#
#  formatOutput
#
#  This takes a string representation of the result and formats it according
#  to a whole bunch of options.
#
#******************************************************************************

def formatOutput( output ):
    # filter out text strings
    for c in output:
        if c in '+-.':
            continue

        # anything with embedded whitespace is a string
        if c in string.whitespace or c in string.punctuation:
            return output

    # output settings, which may be overrided by temp settings
    outputRadix = g.outputRadix
    integerGrouping = g.integerGrouping
    leadingZero = g.leadingZero
    integerDelimiter = g.integerDelimiter

    # override settings with temporary settings if needed
    if g.tempCommaMode:
        integerGrouping = 3     # override whatever was set on the command-line
        leadingZero = False     # this one, too
        integerDelimiter = ','

    if g.tempHexMode:
        integerGrouping = 4
        leadingZero = True
        outputRadix = 16

    if g.tempLeadingZeroMode:
        leadingZero = True

    if g.tempOctalMode:
        integerGrouping = 3
        leadingZero = True
        outputRadix = 8

    mpOutput = mpmathify( output )

    imaginary = im( mpOutput )
    real = re( mpOutput )

    result, negative = formatNumber( real, outputRadix, leadingZero, integerGrouping,
                                     integerDelimiter, g.decimalDelimiter )

    if negative:
        result = '-' + result

    if imaginary != 0:
        strImaginary, negativeImaginary = \
            formatNumber( imaginary, outputRadix, leadingZero, integerGrouping, integerDelimiter, g.decimalDelimiter )
        result = '( ' + result + ( ' - ' if negativeImaginary else ' + ' ) + strImaginary + 'j )'

    #print( 'formatOutput 2:', output )
    #print( )

    return result


#******************************************************************************
#
#  formatListOutput
#
#******************************************************************************

def formatListOutput( result, level=0, indent=0, file=sys.stdout ):
    '''
    This function formats a list for output, taking into account the list
    format level, as specified by the -s option.
    '''
    indentString = ' ' * indent

    first = True

    if level < g.listFormatLevel:
        useIndent = True
        print( indentString + '[', end='', file=file )
        levelIndent = ' ' * ( level + 1 ) * 4
    else:
        useIndent = False
        print( indentString + '[ ', end='', file=file )

    #print( 'result type', type( result ) )

    for item in result:
        newString = ''

        if isinstance( item, ( list, RPNGenerator ) ):
            if first:
                first = False

                if useIndent:
                    print( file=file )
                    print( indentString + levelIndent, end='', file=file )
            else:
                if useIndent:
                    print( ',', file=file )
                    print( indentString + levelIndent, end='', file=file )
                else:
                    print( ', ', end='', file=file )

            formatListOutput( item, level + 1, file=file )
            continue

        if isinstance( item, str ):
            newString = item
        elif isinstance( item, RPNDateTime ):
            newString = formatDateTime( item )
        elif isinstance( item, RPNMeasurement ):
            newString = formatOutput( nstr( item.value, min_fixed=-g.maximumFixed - 1 ) )
            newString += ' ' + formatUnits( item )
        else:
            newString = formatOutput( str( item ) )

        if first:
            first = False
        else:
            print( ', ', end='', file=file )

        if useIndent:
            print( file=file )
            print( indentString + levelIndent + newString, end='', flush=True, file=file )
        else:
            print( newString, end='', flush=True, file=file )

    if useIndent:
        print( file=file )
        print( ' ' * level * 4 + ']', end='', file=file )
    else:
        print( ' ]', end='', file=file )

    if level == 0:
        print( file=file )


#******************************************************************************
#
#  formatUnits
#
#  For outputting an RPNMeasurement object, this method formats the units part
#  of the object, and returns the formatted string.
#
#******************************************************************************

def formatUnits( measurement ):
    value = measurement.value

    unitName = measurement.getUnitName( )

    if unitName is not None:
        unitString = ''

        if mpf( -1.0 ) < value > mpf( 1.0 ) or value == 0:
            tempString = measurement.getPluralUnitName( )
        else:
            tempString = unitName

        tempString = tempString.replace( '_null_unit*', '' )
        tempString = tempString.replace( '*_null_unit', '' )

        tempString = tempString.replace( '_null_unit/', '1/' )
        tempString = tempString.replace( '/_null_unit', '' )

        for c in tempString:
            if c == '_':
                unitString += ' '
            else:
                unitString += c

        return unitString

    unitString = ''

    units = measurement.units

    # now that we've expanded the compound units, let's format...
    for unit in units:
        if unit == '_null_unit':
            continue

        exponent = units[ unit ]

        if exponent > 0:
            if unitString != '':
                unitString += ' '

            if value == 1:
                unitString += unit
            else:
                unitString += g.unitOperators[ unit ].plural

            if exponent > 1:
                unitString += '^' + str( int( exponent ) )

    negativeUnits = ''

    if unitString == '':
        for unit in units:
            if unit == '_null_unit':
                continue

            exponent = units[ unit ]

            if exponent < 0:
                if negativeUnits != '':
                    negativeUnits += ' '

                negativeUnits += unit
                negativeUnits += '^' + str( int( exponent ) )
    else:
        for unit in units:
            if unit == '_null_unit':
                continue

            exponent = units[ unit ]

            if exponent < 0:
                if negativeUnits == '':
                    negativeUnits = ' per '
                else:
                    negativeUnits += ' '

                negativeUnits += unit

                if exponent > 1:
                    negativeUnits += '^' + str( int( exponent ) )
                elif exponent < -1:
                    negativeUnits += '^' + str( int( -exponent ) )

    result = ''

    # filter out the underscores
    for c in unitString + negativeUnits:
        if c == '_':
            result += ' '
        else:
            result += c

    return result


#******************************************************************************
#
#  formatDateTime
#
#******************************************************************************

def formatDateTime( datetime ):
    if not isinstance( datetime, RPNDateTime ):
        raise ValueError( 'expected RPNDateTime' )

    if datetime.getDateOnly( ):
        return datetime.formatDate( )

    # if datetime.microsecond:
    #     return datetime.format( 'YYYY-MM-DD HH:mm:ss.SSSSSS' )
    # else:
    return datetime.format( )


#******************************************************************************
#
#  printParagraph
#
#******************************************************************************

def printParagraph( text, indent = 0 ):
    lines = textwrap.wrap( text, g.lineLength - ( indent + 1 ) )

    for line in lines:
        print( ' ' * indent + line )


#******************************************************************************
#
#  printOperatorHelp
#
#******************************************************************************

def printOperatorHelp( term, operatorInfo, operatorHelp, regularOperator = True):
    if regularOperator:
        if operatorInfo.argCount == 1:
            print( 'n ', end = '' )
        elif operatorInfo.argCount == 2:
            print( 'n k ', end = '' )
        elif operatorInfo.argCount == 3:
            print( 'a b c ', end = '' )
        elif operatorInfo.argCount == 4:
            print( 'a b c d ', end = '' )
        elif operatorInfo.argCount == 5:
            print( 'a b c d e ', end = '' )

    aliasList = [ key for key in g.aliases if term == g.aliases[ key ] ]

    print( term + ' - ' + operatorHelp[ 1 ] )

    print( )

    if len( aliasList ) > 1:
        printParagraph( 'aliases:  ' + ', '.join( aliasList ) )
    elif len( aliasList ) == 1:
        printParagraph( 'alias:  ' + aliasList[ 0 ] )

    print( 'category: ' + operatorHelp[ 0 ] )

    if operatorHelp[ 2 ] == '' or operatorHelp[ 2 ] == '\n':
        print( )
        print( 'No further help is available.' )
        print( )
    else:
        print( operatorHelp[ 2 ] )

    if regularOperator:
        if len( operatorHelp ) > 3:
            if operatorHelp[ 3 ] == '' or operatorHelp[ 3 ] == '\n':
                print( 'No examples are available.' )
            else:
                print( term + ' examples:' )
                print( operatorHelp[ 3 ] )
        else:
            print( 'No examples are available.' )

    if len( operatorHelp ) > 4 and len( operatorHelp[ 4 ] ) > 0:
        print( 'see also:  ', end='' )

        bFirst = True

        for name in sorted( operatorHelp[ 4 ] ):
            if bFirst:
                bFirst = False
            else:
                print( ', ', end='' )

            print( name, end='' )

        print( )


#******************************************************************************
#
#  printCategoryHelp
#
#******************************************************************************

def printCategoryHelp( category, operators, listOperators, modifiers, operatorHelp ):
    if category in basicUnitTypes:
        units = [ ]

        for unit, unitInfo in g.unitOperators.items( ):
            if unitInfo.unitType == category:
                units.append( unit )

        printParagraph( 'The ' + category + ' category includes the following units:' )
        print( )

        printParagraph( ', '.join( sorted( units ) ), indent=4 )
        return

    printParagraph( 'The ' + category + ' category includes the following operators (with aliases in parentheses):' )
    print( )

    operatorList = [ key for key in operators if operatorHelp[ key ][ 0 ] == category ]
    operatorList.extend( [ key for key in listOperators if operatorHelp[ key ][ 0 ] == category ] )
    operatorList.extend( [ key for key in modifiers if operatorHelp[ key ][ 0 ] == category ] )

    addAliases( operatorList, g.aliases )

    for operator in sorted( operatorList ):
        print( operator )


#******************************************************************************
#
#  printHelp
#
#******************************************************************************

def printHelp( terms = None, interactive = False ):
    from rpn.rpnOperators import constants, listOperators, modifiers, operators

    if terms is None:
        terms = [ ]

    loadHelpData( )
    loadUnitData( )

    if g.helpVersion != PROGRAM_VERSION:
        print( 'rpn:  help file version mismatch' )

    unitType = False

    if len( terms ) > 0:
        term = terms[ 0 ]
    else:
        term = ''

    if len( terms ) > 1 and terms[ 0 ] == 'unit_type':
        term = terms[ 1 ]
        unitType = True

    if term == '':
        if interactive:
            printInteractiveHelp( )
        else:
            printGeneralHelp( )
        return

    # first check if the term is an alias and translate
    if term in g.aliases:
        term = g.aliases[ term ]

    # then look for exact matches in all the lists of terms for which we have help support
    if term in operators and not unitType:
        printOperatorHelp( term, operators[ term ], g.OPERATOR_HELP[ term ] )
    elif term in g.unitOperators:
        printOperatorHelp( term, g.unitOperators[ term ], g.OPERATOR_HELP[ term ], regularOperator = False )
    elif term in g.constantOperators:
        printOperatorHelp( term, g.constantOperators[ term ], g.OPERATOR_HELP[ term ], regularOperator = False )
    elif term in constants:
        printOperatorHelp( term, constants[ term ], g.OPERATOR_HELP[ term ] )
    elif term in listOperators:
        printOperatorHelp( term, listOperators[ term ], g.OPERATOR_HELP[ term ] )
    elif term in modifiers:
        printOperatorHelp( term, modifiers[ term ], g.OPERATOR_HELP[ term ] )
    elif term in g.HELP_TOPICS:
        print( g.HELP_TOPICS[ term ] )
    elif term in g.operatorCategories:
        printCategoryHelp( term, operators, listOperators, modifiers, g.OPERATOR_HELP )
    elif term == 'unit_types':
        printParagraph( ', '.join( sorted( [ key for key in g.unitTypeDict.keys( ) if key != '_null_type' ] ) ),
                        indent=4 )
    elif term in g.unitTypeDict:
        unitList = sorted( g.unitTypeDict[ term ] )
        addAliases( unitList, g.aliases )
        for unit in unitList:
            printParagraph( unit, indent=4 )
    else:
        # if no exact matches for any topic, let's look for partial matches
        if 'unit_types'.startswith( term ):
            print( 'Interpreting topic as \'unit_types\'.' )
            printParagraph( ', '.join( sorted( g.unitTypeDict.keys( ) ) ), indent=4 )
            return

        helpTerm = next( ( i for i in g.unitTypeDict if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( )
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            printParagraph( ', '.join( sorted( g.unitTypeDict[ helpTerm ] ) ), indent=4 )
            return

        helpTerm = next( ( i for i in operators if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            printOperatorHelp( helpTerm, operators[ helpTerm ], g.OPERATOR_HELP[ helpTerm ] )
            return

        helpTerm = next( ( i for i in constants if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            printOperatorHelp( helpTerm, constants[ helpTerm ], g.OPERATOR_HELP[ helpTerm ] )
            return

        helpTerm = next( ( i for i in listOperators if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            printOperatorHelp( helpTerm, listOperators[ helpTerm ], g.OPERATOR_HELP[ helpTerm ] )
            return

        helpTerm = next( ( i for i in modifiers if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            printOperatorHelp( helpTerm, modifiers[ helpTerm ], g.OPERATOR_HELP[ helpTerm ] )
            return

        helpTerm = next( ( i for i in g.HELP_TOPICS if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            print( g.HELP_TOPICS[ helpTerm ] )
            return

        helpTerm = next( ( i for i in g.operatorCategories if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            printCategoryHelp( helpTerm, operators, listOperators, modifiers, g.OPERATOR_HELP )
        else:
            print( 'Help topic not found.' )


#******************************************************************************
#
#  printGeneralHelp
#
#******************************************************************************

def printGeneralHelp( ):
    print( RPN_PROGRAM_NAME + ' - ' + PROGRAM_DESCRIPTION )
    print( COPYRIGHT_MESSAGE )
    print( )

    printParagraph(
        'For help on a specific topic, use \'rpn help\' and add a help topic, '
        'operator category or a specific operator name.' )

    printHelpTopics( )


#******************************************************************************
#
#  printInteractiveHelp
#
#******************************************************************************

def printInteractiveHelp( ):
    loadHelpData( )

    printParagraph(
        'For help on a specific topic, use the topic operator with a general topic, '
        'operator category or a specific operator name.' )

    printHelpTopics( )


#******************************************************************************
#
#  printHelpTopics
#
#******************************************************************************

def printHelpTopics( ):
    print( )
    print( 'The following is a list of general topics:' )
    print( )

    helpTopics = list( g.HELP_TOPICS.keys( ) )
    helpTopics.append( 'unit_types' )

    printParagraph( ', '.join( sorted( helpTopics ) ), indent=4 )

    print( )
    print( 'The following is a list of operator categories:' )
    print( )

    operatorCategories = set( g.operatorCategories )

    unitTypes = dict( basicUnitTypes )

    for i in unitTypes:
        if i in operatorCategories:
            operatorCategories.remove( i )

    printParagraph( ', '.join( sorted( operatorCategories ) ), indent=4 )

    print( )
    print( 'The following is a list of unit categories\n(use \'rpn help unit_type name\' for more info):' )
    print( )

    del unitTypes[ '_null_type' ]

    printParagraph( ', '.join( sorted( unitTypes ) ), indent=4 )


#******************************************************************************
#
#  printHelpModeHelp
#
#******************************************************************************

def printHelpModeHelp( ):
    printParagraph( 'rpn help mode - \'topics\' for a list of topics, \'exit\' to return to rpn' )


#******************************************************************************
#
#  printTitleScreen
#
#******************************************************************************

def printTitleScreen( programName, programDescription, showHelp=True ):
    print( programName + PROGRAM_VERSION_STRING + ' - ' + programDescription )
    print( COPYRIGHT_MESSAGE )

    if showHelp:
        print( )
        print( 'Type "help" for more information, and "exit" to exit.' )



#!/usr/bin/env python

#*******************************************************************************
#
#  rpnBase.py
#
#  rpnChilada base conversion functions
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#*******************************************************************************

from mpmath import fdiv, floor, fmod, fmul, fneg, fsub, mpmathify

import gmpy2

from rpn.rpnUtils import twoArgFunctionEvaluator
from rpn.rpnValidator import argValidator, IntValidator

import rpn.rpnGlobals as g


#******************************************************************************
#
#  convertToBaseN
#
#******************************************************************************

def convertToBaseN( value, base, outputBaseDigits=False, numerals=g.defaultNumerals ):
    '''
    This handles any integer base as long as there is a big-enough list of
    numerals to use.  In practice this ends up being 0-9, a-z, and A-Z, which
    allows us to support up to base 62.
    '''
    if outputBaseDigits:
        if base < 2:
            raise ValueError( 'base must be greater than 1' )

        if value < 0:
            raise ValueError( '\'get_base_k_digits\' does not support negative numbers.' )
    else:
        if not 2 <= base <= len( numerals ):
            raise ValueError( 'base must be from 2 to {0}'.format( len( numerals ) ) )

    if value == 0:
        return 0

    if value < 0:
        return '-' + convertToBaseN( fneg( value ), base, outputBaseDigits, numerals )

    if outputBaseDigits:
        result = [ ]
    else:
        result = ''

    leftDigits = mpmathify( value )

    while leftDigits > 0:
        modulo = fmod( leftDigits, base )

        if outputBaseDigits:
            result.insert( 0, int( modulo ) )
        else:
            result = numerals[ int( modulo ) ] + result

        leftDigits = floor( fdiv( leftDigits, base ) )

    return result


#******************************************************************************
#
#  convertFractionToBaseN
#
#******************************************************************************

def convertFractionToBaseN( value, base, precision, outputBaseDigits ):
    if outputBaseDigits:
        if base < 2:
            raise ValueError( 'base must be greater than 1' )
    else:
        if not 2 <= base <= len( g.numerals ):
            raise ValueError( 'base must be from 2 to %d' % len( g.numerals ) )

    if value < 0 or value >= 1.0:
        raise ValueError( 'value (%s) must be >= 0 and < 1.0' % value )

    if base == 10:
        return str( value )

    if outputBaseDigits:
        result = [ ]
    else:
        result = ''

    while value > 0 and precision > 0:
        value = fmul( value, base )

        digit = int( value )

        if len( result ) == g.outputAccuracy:
            if digit >= base // 2:
                digit += 1

            break

        if outputBaseDigits:
            result.append( digit )
        else:
            result += g.numerals[ digit ]

        value = fsub( value, digit )
        precision -= 1

    return result


#******************************************************************************
#
#  getBaseKDigitList
#
#  gmpy2's digits( ) function is way faster, but only works for bases 2 to
#  62.  I need to overhaul convertToBaseN( )... it's way too slow.
#
#******************************************************************************

def getBaseKDigitList( n, base, dropZeroes = False ):
    if 1 < base < 63:
        if n < 0:
            raise ValueError( '\'get_base_k_digits\' does not support negative numbers.' )

        asciiDigits = gmpy2.digits( int( n ), int( base ) )

        digits = [ ]

        ord0 = ord( '0' )
        orda = ord( 'a' )
        ordA = ord( 'A' )

        for i in asciiDigits:
            if '0' <= i <= '9':
                digits.append( ord( i ) - ord0 )
            elif 'a' <= i <= 'z':
                digits.append( ord( i ) - orda + 10 )
            else:
                digits.append( ord( i ) - ordA + 36 )
    else:
        digits = convertToBaseN( n, base, outputBaseDigits=True )

    result = [ ]

    for digit in digits:
        if dropZeroes and digit == 0:
            continue

        result.append( digit )

    return result


def getBaseKDigits( n, k ):
    return getBaseKDigitList( n, k )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 2 ) ] )
def getBaseKDigitsOperator( n, k ):
    return getBaseKDigitList( n, k )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 2 ) ] )
def getNonzeroBaseKDigitsOperator( n, k ):
    return getBaseKDigitList( n, k, dropZeroes = True )



#!/usr/bin/env python

#******************************************************************************
#
#  rpnComputer.py
#
#  rpnChilada computing operators
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import struct

from mpmath import fadd, fdiv, floor, fmod, fmul, fsub, fsum, log, mpf, mp, mpmathify, power

from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurementClass import RPNMeasurement
from rpn.rpnSettings import setAccuracy
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator
from rpn.rpnValidator import argValidator, IntValidator, ListValidator, RealValidator

import rpn.rpnGlobals as g


#******************************************************************************
#
#  convertToSignedInt
#
#  two's compliment logic is in effect here
#
#******************************************************************************

def convertToSignedInt( n, k ):
    newPrecision = ( int( k ) * 3 / 10 ) + 3

    if mp.dps < newPrecision:
        setAccuracy( newPrecision )

    value = fadd( n, ( power( 2, fsub( k, 1 ) ) ) )
    value = fmod( value, power( 2, k ) )
    value = fsub( value, ( power( 2, fsub( k, 1 ) ) ) )

    return value


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 1 ) ] )
def convertToSignedIntOperator( n, k ):
    return convertToSignedInt( n, k )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertToCharOperator( n ):
    return convertToSignedInt( n, 8 )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertToShortOperator( n ):
    return convertToSignedInt( n, 16 )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertToLongOperator( n ):
    return convertToSignedInt( n, 32 )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertToLongLongOperator( n ):
    return convertToSignedInt( n, 64 )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertToQuadLongOperator( n ):
    return convertToSignedInt( n, 128 )


#******************************************************************************
#
#  getInvertedBitsOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def getInvertedBits( n ):
    # determine how many groups of bits we will be looking at
    if n == 0:
        groupings = 1
    else:
        groupings = int( fadd( floor( fdiv( ( log( n, 2 ) ), g.bitwiseGroupSize ) ), 1 ) )

    placeValue = mpmathify( 1 << g.bitwiseGroupSize )
    multiplier = mpmathify( 1 )
    remaining = n

    result = mpmathify( 0 )

    for _ in range( 0, groupings ):
        # Let's let Python do the actual inverting
        group = fmod( ~int( fmod( remaining, placeValue ) ), placeValue )

        result += fmul( group, multiplier )

        remaining = floor( fdiv( remaining, placeValue ) )
        multiplier = fmul( multiplier, placeValue )

    return result


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def getInvertedBitsOperator( n ):
    return getInvertedBits( n )


#******************************************************************************
#
#  performBitwiseOperation
#
#  The operations are performed on groups of bits as specified by the variable
#  bitwiseGroupSize.  Although doing it this way isn't really necessary, it
#  does mean that under normal circumstances the regular Python bit operators
#  can be used.
#
#******************************************************************************

def performBitwiseOperation( i, j, operation ):
    value1 = int( i )
    value2 = int( j )

    # determine how many groups of bits we will be looking at
    if value1 == 0:
        groupings = 1
    else:
        groupings = int( fadd( floor( fdiv( ( log( value1, 2 ) ), g.bitwiseGroupSize ) ), 1 ) )

    if value2 == 0:
        groupings2 = 1
    else:
        groupings2 = int( fadd( floor( fdiv( ( log( value2, 2 ) ), g.bitwiseGroupSize ) ), 1 ) )

    if groupings2 > groupings:
        groupings = groupings2

    placeValue = mpmathify( 1 << g.bitwiseGroupSize )
    multiplier = mpmathify( 1 )
    remaining1 = value1
    remaining2 = value2

    result = mpmathify( 0 )

    for _ in range( 0, groupings ):
        mod1 = fmod( remaining1, placeValue )
        mod2 = fmod( remaining2, placeValue )

        result = fadd( fmul( operation( int( mod1 ), int( mod2 ) ), multiplier ), result )

        remaining1 = floor( fdiv( remaining1, placeValue ) )
        remaining2 = floor( fdiv( remaining2, placeValue ) )

        multiplier = fmul( multiplier, placeValue )

    return result


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def getBitwiseAndOperator( n, k ):
    return performBitwiseOperation( n, k, lambda x, y: x & y )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def getBitwiseNandOperator( n, k ):
    return getInvertedBits( performBitwiseOperation( n, k, lambda x, y: x & y ) )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def getBitwiseNorOperator( n, k ):
    return getInvertedBits( performBitwiseOperation( n, k, lambda x, y: x | y ) )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def getBitwiseOrOperator( n, k ):
    return performBitwiseOperation( n, k, lambda x, y: x | y )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def getBitwiseXorOperator( n, k ):
    return performBitwiseOperation( n, k, lambda x, y: x ^ y )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def getBitwiseXnorOperator( n, k ):
    return getInvertedBits( performBitwiseOperation( n, k, lambda x, y: x ^ y ) )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def shiftLeftOperator( n, k ):
    return fmul( n, 1 << int( k ) )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def shiftRightOperator( n, k ):
    return floor( fdiv( n, 1 << int( k ) ) )


#******************************************************************************
#
#  getBitCount
#
#******************************************************************************

def getBitCount( n ):
    n = int( n )

    if n < 0:
        raise ValueError( '\'bit_count\' requires a positive integer value' )

    result = 0

    if isinstance( n, RPNMeasurement ):
        value = n.value
    else:
        value = n

    while value:
        result += fmod( value, 2 )
        value = floor( fdiv( value, 2 ) )

    return result


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def getBitCountOperator( n ):
    return getBitCount( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def getParityOperator( n ):
    return fmod( getBitCount( n ), 2 )


#******************************************************************************
#
#  unpackInteger
#
#******************************************************************************

def unpackInteger( n, fields ):
    if isinstance( n, RPNGenerator ):
        return unpackInteger( list( n ), fields )

    if isinstance( n, list ):
        return [ unpackInteger( i, fields ) for i in n ]

    if isinstance( fields, RPNGenerator ):
        return unpackInteger( n, list( fields ) )

    if not isinstance( fields, list ):
        return unpackInteger( n, [ fields ] )

    value = int( n )
    result = [ ]

    for field in reversed( fields ):
        size = int( field )
        result.insert( 0, value & ( 2 ** size - 1 ) )
        value >>= size

    return result


#oneArgAndListFunctionEvaluator( )
@argValidator( [ IntValidator( ), ListValidator( ) ] )
def unpackIntegerOperator( n, fields ):
    return unpackInteger( n, fields )


#******************************************************************************
#
#  packInteger
#
#******************************************************************************

def packInteger( values, fields ):
    if isinstance( values, RPNGenerator ):
        return packInteger( list( values ), fields )

    if not isinstance( values, list ):
        return unpackInteger( [ values ], fields )

    if isinstance( fields, RPNGenerator ):
        return packInteger( values, list( fields ) )

    if not isinstance( fields, list ):
        return unpackInteger( values, [ fields ] )

    if isinstance( values[ 0 ], list ):
        return [ unpackInteger( value, fields ) for value in values ]

    result = 0

    count = min( len( values ), len( fields ) )

    size = 0

    for i in range( count, 0, -1 ):
        field = int( fields[ i - 1 ] )
        value = int( values[ i - 1 ] ) & ( 2 ** field - 1 )
        result = fadd( result, fmul( value, power( 2, size ) ) )
        size += field

    return result


@argValidator( [ ListValidator( ), ListValidator( ) ] )
def packIntegerOperator( values, fields ):
    return packInteger( values, fields )


#******************************************************************************
#
#  interpretAsFloat
#
#******************************************************************************

def interpretAsFloat( n ):
    intValue = struct.pack( 'I', int( n ) )
    return mpf( struct.unpack( 'f', intValue )[ 0 ] )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0, 2 ** 32 - 1 ) ] )
def interpretAsFloatOperator( n ):
    return interpretAsFloat( n )


#******************************************************************************
#
#  interpretAsDouble
#
#******************************************************************************

def interpretAsDouble( n ):
    setAccuracy( 25 )

    intValue = struct.pack( 'Q', int( n ) )
    return mpf( struct.unpack( 'd', intValue )[ 0 ] )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0, 2 ** 64 - 1 ) ] )
def interpretAsDoubleOperator( n ):
    return interpretAsDouble( n )


#******************************************************************************
#
#  conversion operators
#
#******************************************************************************


@oneArgFunctionEvaluator( )
@argValidator( [ RealValidator( ) ] )
def convertToDoubleOperator( n ):
    return fsum( b << 8 * i for i, b in enumerate( struct.pack( 'd', float( n ) ) ) )


@oneArgFunctionEvaluator( )
@argValidator( [ RealValidator( ) ] )
def convertToFloatOperator( n ):
    return fsum( b << 8 * i for i, b in enumerate( struct.pack( 'f', float( n ) ) ) )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( 1 ) ] )
def convertToUnsignedIntOperator( n, k ):
    return fmod( n, power( 2, k ) )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertToUnsignedCharOperator( n ):
    return fmod( n, power( 2, 8 ) )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertToUnsignedShortOperator( n ):
    return fmod( n, power( 2, 16 ) )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertToUnsignedLongOperator( n ):
    return fmod( n, power( 2, 32 ) )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertToUnsignedLongLongOperator( n ):
    return fmod( n, power( 2, 64 ) )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def convertToUnsignedQuadLongOperator( n ):
    return fmod( n, power( 2, 128 ) )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def andOperator( n, k ):
    return 1 if ( n != 0 and k != 0 ) else 0


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def notOperator( n ):
    return 1 if n == 0 else 0


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def nandOperator( n, k ):
    return 0 if ( n != 0 and k != 0 ) else 1


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def norOperator( n, k ):
    return 0 if ( n != 0 or k != 0 ) else 1


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def orOperator( n, k ):
    return 1 if ( n != 0 or k != 0 ) else 0


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def xnorOperator( n, k ):
    return 1 if ( n != 0 ) == ( k != 0 ) else 0


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( ), IntValidator( ) ] )
def xorOperator( n, k ):
    return 1 if ( n != 0 ) != ( k != 0 ) else 0



#!/usr/bin/env python

#******************************************************************************
#
#  rpnCalendar.py
#
#  rpnChilada calendar operators
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import calendar

from convertdate import bahai, french_republican, hebrew, indian_civil, islamic, julian, \
                        julianday, mayan, persian
from ethiopian_date import EthiopianDateConverter as ethiopian_date
from mpmath import ceil

from rpn.rpnDateTime import RPNDateTime
from rpn.rpnName import getOrdinalName
from rpn.rpnUtils import oneArgFunctionEvaluator
from rpn.rpnValidator import argValidator, DateTimeValidator, IntValidator, YearValidator

#******************************************************************************
#
#  calendar names
#
#******************************************************************************

bahaiYears = [
    'Alif',
    'Ba\'',
    'Ab',
    'Dal',
    'Bab',
    'Vav',
    'Abad',
    'Jad',
    'Baha\'',
    'Hubb',
    'Bahhaj',
    'Javab',
    'Ahad',
    'Vahhab',
    'Vidad',
    'Badi`',
    'Bahi',
    'Abha',
    'Vahid',
]

bahaiMonths = [
    'Baha',
    'Jalal',
    'Jamal',
    '`Azamat',
    'Nur',
    'Rahmat',
    'Kalimat',
    'Kamal',
    'Asma\'',
    '`Izzat',
    'Mashiyyat',
    '`Ilm',
    'Qudrat',
    'Qawl',
    'Masa\'il',
    'Sharaf',
    'Sultan',
    'Mulk',
    'Ayyam-i-Ha',
    '`Ala\''
]

bahaiDays = [
    'Kamal',     # Monday
    'Fidal',
    '`Idal',
    'Istijlal',
    'Istiqlal',
    'Jalal',
    'Jamal'
]

hebrewMonths = [
    'Nisan',
    'Iyar',
    'Sivan',
    'Tammuz',
    'Av',
    'Elul',
    'Tishrei',
    'Marcheshvan',
    'Kislev',
    'Tevet',
    'Shevat',
    'Adar I',
    'Adar II'
]

hebrewDays = [
    'Yom Sheni',     # Monday
    'Yom Shlishi',
    'Yom Revi\'i',
    'Yom Chamishi',
    'Yom Shishi',
    'Yom Shabbat',
    'Yom Rishon'
]

indianCivilDays = [
    'Somavara',     # Monday
    'Mangalavara',
    'Budhavara',
    'Guruvara',
    'Sukravara',
    'Sanivara',
    'Ravivara',
]

indianCivilMonths = [
    'Chaitra',
    'Vaishakha',
    'Jyeshtha',
    'Ashadha',
    'Shravana',
    'Bhaadra',
    'Ashwin',
    'Kartika',
    'Agrahayana',
    'Pausha',
    'Magha',
    'Phalguna',
]

islamicMonths = [
    'Muharram',
    'Safar',
    'Rabi al-Awwal',
    'Rabi ath-Thani',
    'Jumada al-Ula',
    'Jumada ath-Thaniyah',
    'Rajab',
    'Sha`ban',
    'Ramadan',
    'Shawwal',
    'Dhu al-Qa`dah',
    'Dhu al-Hijjah'
]

islamicDays = [
    'al-Ithnayn',   # Monday
    'ath-Thulatha',
    'al-Arbi`a',
    'al-Khamis',
    'al-Jumu`ah',
    'as-Sabt',
    'al-Ahad'
]

persianMonths = [
    'Farvardin',
    'Ordibehesht',
    'Khordad',
    'Tir',
    'Mordad',
    'Shahrivar',
    'Mehr',
    'Aban',
    'Azar',
    'Dey',
    'Bahman',
    'Esfand'
]

persianDays = [
    'Yekshanbeh',   # Monday
    'Doshanbeh',
    'Seshhanbeh',
    'Chaharshanbeh',
    'Panjshanbeh',
    'Jomeh',
    'Shanbeh'
]

ethiopianMonths = [
    'Meskerem',
    'Thikimt',
    'Hidar',
    'Tahsas',
    'Thir',
    'Yekatit',
    'Megabit',
    'Miyazya',
    'Ginbot',
    'Senie',
    'Hamlie',
    'Nehasie',
    'Phagumien'
]

ethiopianDays = [
    'Lideta',
    'Aba Guba',
    'Be\'Eta',
    'Yohannes',
    'Abo',
    'Iyesus',
    'Selassie',
    'Aba Kiros',
    'Tomas',
    'Meskel',
    'Hana Mariam',
    'Michael',
    'Egziher Ab',
    'Abune Aregawi',
    'K\'irk\'os',
    'Kidane Mihret',
    'Est\'ifanos',
    'Tekle Alfa',
    'Gabriel',
    'Hintsite',
    'Mariam',
    'Ura\'el',
    'Giorgis',
    'Tekle Haimanot',
    'Merk\'orios',
    'Yosef',
    'Medhani Alem',
    'Amanu\'el',
    'Bale Egziabeir',
    'Mark\'os'
]

frenchRepublicanDays = [
    'Primidi',
    'Duodi',
    'Tridi',
    'Quartidi',
    'Quintidi',
    'Sextidi',
    'Septidi',
    'Octidi',
    'Nonidi',
    'Decadi'
]

frenchRepublicanMonths = [
    'Vendemiaire',
    'Brumaire',
    'Frimaire',
    'Nivose',
    'Pluviose',
    'Ventose',
    'Germinal',
    'Floreal',
    'Prairial',
    'Messidor',
    'Thermidor',
    'Fructidor',
    'Sansculottides'
]


#******************************************************************************
#
#  getOrdinalDateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getOrdinalDateOperator( n ):
    return str( n.year ) + '-' + str( n.timetuple( ).tm_yday )


#******************************************************************************
#
#  generateMonthCalendarOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def generateMonthCalendarOperator( datetime ):
    cal = calendar.TextCalendar( firstweekday = 6 )
    print( )
    cal.prmonth( datetime.year, datetime.month )
    print( )

    return [ datetime.year, datetime.month ]


#******************************************************************************
#
#  generateYearCalendarOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ YearValidator( ) ] )
def generateYearCalendarOperator( n ):
    cal = calendar.TextCalendar( firstweekday = 6 )

    if isinstance( n, RPNDateTime ):
        cal.pryear( n.year )
    else:
        cal.pryear( int( n ) )

    return ''


#******************************************************************************
#
#  getJulianDayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getJulianDayOperator( n ):
    n.to( 'utc' )
    return julianday.from_datetime( n )


#******************************************************************************
#
#  getLilianDayOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getLilianDayOperator( n ):
    return ceil( n.subtract( RPNDateTime( 1582, 10, 15 ) ).value )


#******************************************************************************
#
#  getISODateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getISODateOperator( n ):
    result = n.isocalendar( )

    return list( result )


#******************************************************************************
#
#  getISODateNameOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getISODateNameOperator( n ):
    result = n.isocalendar( )

    return str( result[ 0 ] ) + '-W' + str( result[ 1 ] ) + '-' + str( result[ 2 ] )


#******************************************************************************
#
#  getHebrewCalendarDateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getHebrewCalendarDateOperator( n ):
    return list( hebrew.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertHebrewDateOperator
#
#******************************************************************************

@argValidator( [ IntValidator( 1 ), IntValidator( 1, 13 ), IntValidator( 1, 30 ) ] )
def convertHebrewDateOperator( year, month, day ):
    return RPNDateTime( *hebrew.to_gregorian( int( year ), int( month ), int( day ) ), dateOnly = True )


#******************************************************************************
#
#  getHebrewCalendarDateNameOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getHebrewCalendarDateNameOperator( n ):
    date = hebrew.from_gregorian( n.year, n.month, n.day )

    return hebrewDays[ n.weekday( ) ] + ', ' + hebrewMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', ' + str( date[ 0 ] )


#******************************************************************************
#
#  getIndianCivilCalendarDateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getIndianCivilCalendarDateOperator( n ):
    return list( indian_civil.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertIndianCivilDateOperator
#
#******************************************************************************

@argValidator( [ IntValidator( 1 ), IntValidator( 1, 12 ), IntValidator( 1, 31 ) ] )
def convertIndianCivilDateOperator( year, month, day ):
    return RPNDateTime( *indian_civil.to_gregorian( int( year ), int( month ), int( day ) ), dateOnly = True )


#******************************************************************************
#
#  getIndianCivilCalendarDateNameOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getIndianCivilCalendarDateNameOperator( n ):
    date = indian_civil.from_gregorian( n.year, n.month, n.day )

    return indianCivilDays[ n.weekday( ) ] + ', ' + indianCivilMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', ' + str( date[ 0 ] )


#******************************************************************************
#
#  getMayanCalendarDateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getMayanCalendarDateOperator( n ):
    return list( mayan.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertMayanDateOperator
#
#******************************************************************************

@argValidator( [ IntValidator( ), IntValidator( 0 ), IntValidator( 0 ), IntValidator( 0 ), IntValidator( 0 ) ] )
def convertMayanDateOperator( baktun, katun, tun, uinal, kin ):
    return RPNDateTime( *mayan.to_gregorian( baktun, katun, tun, uinal, kin ), dateOnly = True )


#******************************************************************************
#
#  getIslamicCalendarDateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getIslamicCalendarDateOperator( n ):
    return list( islamic.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertIslamicDateOperator
#
#******************************************************************************

@argValidator( [ IntValidator( 1 ), IntValidator( 1, 12 ), IntValidator( 1, 31 ) ] )
def convertIslamicDateOperator( year, month, day ):
    return RPNDateTime( *islamic.to_gregorian( int( year ), int( month ), int( day ) ), dateOnly = True )


#******************************************************************************
#
#  getIslamicCalendarDateNameOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getIslamicCalendarDateNameOperator( n ):
    date = islamic.from_gregorian( n.year, n.month, n.day )

    return islamicDays[ n.weekday( ) ] + ', ' + islamicMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', ' + str( date[ 0 ] )


#******************************************************************************
#
#  getJulianCalendarDateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getJulianCalendarDateOperator( n ):
    return list( julian.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertJulianDateOperator
#
#******************************************************************************

@argValidator( [ IntValidator( ), IntValidator( 1, 12 ), IntValidator( 1, 31 ) ] )
def convertJulianDateOperator( year, month, day ):
    return RPNDateTime( *julian.to_gregorian( int( year ), int( month ), int( day ) ), dateOnly = True )


#******************************************************************************
#
#  getPersianCalendarDateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getPersianCalendarDateOperator( n ):
    return list( persian.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertPersianDateOperator
#
#******************************************************************************

@argValidator( [ IntValidator( ), IntValidator( 1, 12 ), IntValidator( 1, 31 ) ] )
def convertPersianDateOperator( year, month, day ):
    return RPNDateTime( *persian.to_gregorian( int( year ), int( month ), int( day ) ), dateOnly = True )


#******************************************************************************
#
#  getPersianCalendarDateNameOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getPersianCalendarDateNameOperator( n ):
    date = persian.from_gregorian( n.year, n.month, n.day )

    return persianDays[ n.weekday( ) ] + ', ' + persianMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', ' + str( date[ 0 ] )


#******************************************************************************
#
#  getBahaiCalendarDateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getBahaiCalendarDateOperator( n ):
    return list( bahai.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertBahaiDateOperator
#
#******************************************************************************

@argValidator( [ IntValidator( ), IntValidator( 1, 19 ), IntValidator( 1, 31 ) ] )
def convertBahaiDateOperator( year, month, day ):
    return RPNDateTime( *bahai.to_gregorian( int( year ), int( month ), int( day ) ), dateOnly = True )


#******************************************************************************
#
#  getBahaiCalendarDateNameOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getBahaiCalendarDateNameOperator( n ):
    date = bahai.from_gregorian( n.year, n.month, n.day )

    result = bahaiDays[ n.weekday( ) ] + ', ' + bahaiMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', '

    if date[ 0 ] >= 1:
        result += 'Year ' + bahaiYears[ date[ 0 ] % 19 - 1 ] + ' of the ' + \
                  getOrdinalName( ( date[ 0 ] // 19 ) + 1 ) + ' Vahid of the ' + \
                  getOrdinalName( ( date[ 0 ] // 361 ) + 1 ) + ' Kull-i-Shay\'' + \
                  ' (Year ' + str( date[ 0 ] ) + ')'
    else:
        result += str( date[ 0 ] )

    return result


#******************************************************************************
#
#  getEthiopianCalendarDateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getEthiopianCalendarDateOperator( n ):
    return list( ethiopian_date.to_ethiopian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertEthiopianDateOperator
#
#******************************************************************************

@argValidator( [ IntValidator( ), IntValidator( 1, 12 ), IntValidator( 1, 31 ) ] )
def convertEthiopianDateOperator( year, month, day ):
    ethDate = ethiopian_date.to_gregorian( int( year ), int( month ), int( day ) )
    return RPNDateTime( ethDate.year, ethDate.month, ethDate.day )


#******************************************************************************
#
#  getEthiopianCalendarDateNameOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getEthiopianCalendarDateNameOperator( n ):
    date = list( ethiopian_date.to_ethiopian( n.year, n.month, n.day ) )

    return ethiopianDays[ date[ 2 ] - 1 ] + ' ' + ethiopianMonths[ date[ 1 ] - 1 ] + \
                          ' ' + str( date[ 0 ] )


#******************************************************************************
#
#  getFrenchRepublicanCalendarDateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getFrenchRepublicanCalendarDateOperator( n ):
    return list( french_republican.from_gregorian( n.year, n.month, n.day ) )


#******************************************************************************
#
#  convertFrenchRepublicanDateOperator
#
#******************************************************************************

@argValidator( [ IntValidator( ), IntValidator( 1, 13 ), IntValidator( 1, 31 ) ] )
def convertFrenchRepublicanDateOperator( year, month, day ):
    return RPNDateTime( *french_republican.to_gregorian( int( year ), int( month ), int( day ) ), dateOnly = True )


#******************************************************************************
#
#  getFrenchRepublicanCalendarDateNameOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ DateTimeValidator( ) ] )
def getFrenchRepublicanCalendarDateNameOperator( n ):
    date = french_republican.from_gregorian( n.year, n.month, n.day )

    #return frenchRepublicanDays[ n.weekday( ) ] + ', ' + frenchRepublicanMonths[ date[ 1 ] - 1 ] + \
    #       ' ' + str( date[ 2 ] ) + ', ' + str( int( date[ 0 ] ) )
    return frenchRepublicanMonths[ date[ 1 ] - 1 ] + \
           ' ' + str( date[ 2 ] ) + ', ' + str( int( date[ 0 ] ) )
